/* Generated by Cython 0.29.26 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "yt/utilities/lib/platform_dep.h"
        ],
        "include_dirs": [
            "yt/utilities/lib"
        ],
        "libraries": [
            "m"
        ],
        "name": "yt.utilities.lib.primitives",
        "sources": [
            "yt/utilities/lib/primitives.pyx"
        ]
    },
    "module_name": "yt.utilities.lib.primitives"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_26"
#define CYTHON_HEX_VERSION 0x001D1AF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030B00A1)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__yt__utilities__lib__primitives
#define __PYX_HAVE_API__yt__utilities__lib__primitives
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include <math.h>
#include "platform_dep.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "yt/utilities/lib/primitives.pyx",
  "__init__.pxd",
  "type.pxd",
};

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray;
struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox;
struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData;
struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle;
struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch;
struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch;

/* "yt/utilities/lib/primitives.pxd":10
 * 
 * 
 * cdef struct Ray:             # <<<<<<<<<<<<<<
 *     np.float64_t origin[3]
 *     np.float64_t direction[3]
 */
struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray {
  __pyx_t_5numpy_float64_t origin[3];
  __pyx_t_5numpy_float64_t direction[3];
  __pyx_t_5numpy_float64_t inv_dir[3];
  __pyx_t_5numpy_float64_t data_val;
  __pyx_t_5numpy_float64_t t_near;
  __pyx_t_5numpy_float64_t t_far;
  __pyx_t_5numpy_int64_t elem_id;
  __pyx_t_5numpy_int64_t near_boundary;
};

/* "yt/utilities/lib/primitives.pxd":20
 *     np.int64_t near_boundary
 * 
 * cdef struct BBox:             # <<<<<<<<<<<<<<
 *     np.float64_t left_edge[3]
 *     np.float64_t right_edge[3]
 */
struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox {
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t right_edge[3];
};

/* "yt/utilities/lib/primitives.pxd":24
 *     np.float64_t right_edge[3]
 * 
 * cdef struct RayHitData:             # <<<<<<<<<<<<<<
 *     np.float64_t u
 *     np.float64_t v
 */
struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData {
  __pyx_t_5numpy_float64_t u;
  __pyx_t_5numpy_float64_t v;
  __pyx_t_5numpy_float64_t t;
  __pyx_t_5numpy_int64_t converged;
};

/* "yt/utilities/lib/primitives.pxd":30
 *     np.int64_t converged
 * 
 * cdef struct Triangle:             # <<<<<<<<<<<<<<
 *     np.float64_t p0[3]
 *     np.float64_t p1[3]
 */
struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle {
  __pyx_t_5numpy_float64_t p0[3];
  __pyx_t_5numpy_float64_t p1[3];
  __pyx_t_5numpy_float64_t p2[3];
  __pyx_t_5numpy_int64_t elem_id;
};

/* "yt/utilities/lib/primitives.pxd":50
 *                         BBox* bbox) nogil
 * 
 * cdef struct Patch:             # <<<<<<<<<<<<<<
 *     np.float64_t[8][3] v  # 8 vertices per patch
 *     np.int64_t elem_id
 */
struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch {
  __pyx_t_5numpy_float64_t v[8][3];
  __pyx_t_5numpy_int64_t elem_id;
};

/* "yt/utilities/lib/primitives.pxd":85
 *                      BBox* bbox) nogil
 * 
 * cdef struct TetPatch:             # <<<<<<<<<<<<<<
 *     np.float64_t[6][3] v # 6 vertices per patch
 *     np.int64_t elem_id
 */
struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch {
  __pyx_t_5numpy_float64_t v[6][3];
  __pyx_t_5numpy_int64_t elem_id;
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'libc.math' */

/* Module declarations from 'yt.utilities.lib.vec3_ops' */
static CYTHON_INLINE float __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(float *, float *); /*proto*/
static CYTHON_INLINE double __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(double *, double *); /*proto*/
static CYTHON_INLINE void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(float *, float *, float *); /*proto*/
static CYTHON_INLINE void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(double *, double *, double *); /*proto*/
static CYTHON_INLINE void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_subtract(double *, double *, double *); /*proto*/
static CYTHON_INLINE float __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(float *, float *); /*proto*/
static CYTHON_INLINE double __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(double *, double *); /*proto*/
static CYTHON_INLINE float __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(float *); /*proto*/
static CYTHON_INLINE double __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(double *); /*proto*/

/* Module declarations from 'yt.utilities.lib.primitives' */
static __pyx_t_5numpy_float64_t __pyx_v_2yt_9utilities_3lib_10primitives_DETERMINANT_EPS;
static __pyx_t_5numpy_float64_t __pyx_v_2yt_9utilities_3lib_10primitives_INF;
static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(float (*)[3], float const , float const , float *); /*proto*/
static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(double (*)[3], double const , double const , double *); /*proto*/
static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU(float (*)[3], float const , float const , float *); /*proto*/
static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU(double (*)[3], double const , double const , double *); /*proto*/
static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV(float (*)[3], float const , float const , float *); /*proto*/
static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV(double (*)[3], double const , double const , double *); /*proto*/
static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit(float (*)[3], float *, float *); /*proto*/
static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit(double (*)[3], double *, double *); /*proto*/
static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit(float (*)[3], float *, float *); /*proto*/
static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit(double (*)[3], double *, double *); /*proto*/
static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(float (*)[3], float const , float const , float *); /*proto*/
static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(double (*)[3], double const , double const , double *); /*proto*/
static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU(float (*)[3], float const , float const , float *); /*proto*/
static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU(double (*)[3], double const , double const , double *); /*proto*/
static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV(float (*)[3], float const , float const , float *); /*proto*/
static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV(double (*)[3], double const , double const , double *); /*proto*/
#define __Pyx_MODULE_NAME "yt.utilities.lib.primitives"
extern int __pyx_module_is_main_yt__utilities__lib__primitives;
int __pyx_module_is_main_yt__utilities__lib__primitives = 0;

/* Implementation of 'yt.utilities.lib.primitives' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_np[] = "np";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_This_file_contains_definitions[] = "\n\nThis file contains definitions of the various primitives that can be used\nby the Cython ray-tracer for unstructured mesh rendering. To define a new\nprimitive type, you need to define a struct that represents it. You also\nneed to provide three functions:\n\n1. A function that computes the intersection between a given ray and a given primitive.\n2. A function that computes the centroid of the primitive type.\n3. A function that computes the axis-aligned bounding box of a given primitive.\n\n";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
/* Late includes */

/* "yt/utilities/lib/primitives.pyx":36
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_bbox_intersect(Ray* ray, const BBox bbox) nogil:             # <<<<<<<<<<<<<<
 *     '''
 * 
 */

static __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_10primitives_ray_bbox_intersect(struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *__pyx_v_ray, struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox const __pyx_v_bbox) {
  __pyx_t_5numpy_float64_t __pyx_v_tmin;
  __pyx_t_5numpy_float64_t __pyx_v_tmax;
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_t1;
  __pyx_t_5numpy_float64_t __pyx_v_t2;
  __pyx_t_5numpy_int64_t __pyx_r;
  __pyx_t_5numpy_int64_t __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":46
 *     # https://tavianator.com/fast-branchless-raybounding-box-intersections/
 * 
 *     cdef np.float64_t tmin = -INF             # <<<<<<<<<<<<<<
 *     cdef np.float64_t tmax =  INF
 * 
 */
  __pyx_v_tmin = (-__pyx_v_2yt_9utilities_3lib_10primitives_INF);

  /* "yt/utilities/lib/primitives.pyx":47
 * 
 *     cdef np.float64_t tmin = -INF
 *     cdef np.float64_t tmax =  INF             # <<<<<<<<<<<<<<
 * 
 *     cdef np.int64_t i
 */
  __pyx_v_tmax = __pyx_v_2yt_9utilities_3lib_10primitives_INF;

  /* "yt/utilities/lib/primitives.pyx":51
 *     cdef np.int64_t i
 *     cdef np.float64_t t1, t2
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         t1 = (bbox.left_edge[i]  - ray.origin[i])*ray.inv_dir[i]
 *         t2 = (bbox.right_edge[i] - ray.origin[i])*ray.inv_dir[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":52
 *     cdef np.float64_t t1, t2
 *     for i in range(3):
 *         t1 = (bbox.left_edge[i]  - ray.origin[i])*ray.inv_dir[i]             # <<<<<<<<<<<<<<
 *         t2 = (bbox.right_edge[i] - ray.origin[i])*ray.inv_dir[i]
 *         tmin = fmax(tmin, fmin(t1, t2))
 */
    __pyx_v_t1 = (((__pyx_v_bbox.left_edge[__pyx_v_i]) - (__pyx_v_ray->origin[__pyx_v_i])) * (__pyx_v_ray->inv_dir[__pyx_v_i]));

    /* "yt/utilities/lib/primitives.pyx":53
 *     for i in range(3):
 *         t1 = (bbox.left_edge[i]  - ray.origin[i])*ray.inv_dir[i]
 *         t2 = (bbox.right_edge[i] - ray.origin[i])*ray.inv_dir[i]             # <<<<<<<<<<<<<<
 *         tmin = fmax(tmin, fmin(t1, t2))
 *         tmax = fmin(tmax, fmax(t1, t2))
 */
    __pyx_v_t2 = (((__pyx_v_bbox.right_edge[__pyx_v_i]) - (__pyx_v_ray->origin[__pyx_v_i])) * (__pyx_v_ray->inv_dir[__pyx_v_i]));

    /* "yt/utilities/lib/primitives.pyx":54
 *         t1 = (bbox.left_edge[i]  - ray.origin[i])*ray.inv_dir[i]
 *         t2 = (bbox.right_edge[i] - ray.origin[i])*ray.inv_dir[i]
 *         tmin = fmax(tmin, fmin(t1, t2))             # <<<<<<<<<<<<<<
 *         tmax = fmin(tmax, fmax(t1, t2))
 * 
 */
    __pyx_v_tmin = fmax(__pyx_v_tmin, fmin(__pyx_v_t1, __pyx_v_t2));

    /* "yt/utilities/lib/primitives.pyx":55
 *         t2 = (bbox.right_edge[i] - ray.origin[i])*ray.inv_dir[i]
 *         tmin = fmax(tmin, fmin(t1, t2))
 *         tmax = fmin(tmax, fmax(t1, t2))             # <<<<<<<<<<<<<<
 * 
 *     return tmax >= fmax(tmin, 0.0)
 */
    __pyx_v_tmax = fmin(__pyx_v_tmax, fmax(__pyx_v_t1, __pyx_v_t2));
  }

  /* "yt/utilities/lib/primitives.pyx":57
 *         tmax = fmin(tmax, fmax(t1, t2))
 * 
 *     return tmax >= fmax(tmin, 0.0)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = (__pyx_v_tmax >= fmax(__pyx_v_tmin, 0.0));
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":36
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_bbox_intersect(Ray* ray, const BBox bbox) nogil:             # <<<<<<<<<<<<<<
 *     '''
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/primitives.pyx":62
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_triangle_intersect(const void* primitives,             # <<<<<<<<<<<<<<
 *                                        const np.int64_t item,
 *                                        Ray* ray) nogil:
 */

static __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_10primitives_ray_triangle_intersect(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *__pyx_v_ray) {
  struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle __pyx_v_tri;
  __pyx_t_5numpy_float64_t __pyx_v_e1[3];
  __pyx_t_5numpy_float64_t __pyx_v_e2[3];
  __pyx_t_5numpy_float64_t __pyx_v_P[3];
  __pyx_t_5numpy_float64_t __pyx_v_det;
  __pyx_t_5numpy_float64_t __pyx_v_inv_det;
  __pyx_t_5numpy_float64_t __pyx_v_T[3];
  __pyx_t_5numpy_float64_t __pyx_v_u;
  __pyx_t_5numpy_float64_t __pyx_v_Q[3];
  __pyx_t_5numpy_float64_t __pyx_v_v;
  __pyx_t_5numpy_float64_t __pyx_v_t;
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":77
 *     # https://en.wikipedia.org/wiki/M%C3%B6ller%E2%80%93Trumbore_intersection_algorithm
 * 
 *     cdef Triangle tri = (<Triangle*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     # edge vectors
 */
  __pyx_v_tri = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":82
 *     cdef np.float64_t e1[3]
 *     cdef np.float64_t e2[3]
 *     subtract(tri.p1, tri.p0, e1)             # <<<<<<<<<<<<<<
 *     subtract(tri.p2, tri.p0, e2)
 * 
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_subtract(__pyx_v_tri.p1, __pyx_v_tri.p0, __pyx_v_e1);

  /* "yt/utilities/lib/primitives.pyx":83
 *     cdef np.float64_t e2[3]
 *     subtract(tri.p1, tri.p0, e1)
 *     subtract(tri.p2, tri.p0, e2)             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t P[3]
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_subtract(__pyx_v_tri.p2, __pyx_v_tri.p0, __pyx_v_e2);

  /* "yt/utilities/lib/primitives.pyx":86
 * 
 *     cdef np.float64_t P[3]
 *     cross(ray.direction, e2, P)             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t det, inv_det
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(__pyx_v_ray->direction, __pyx_v_e2, __pyx_v_P);

  /* "yt/utilities/lib/primitives.pyx":89
 * 
 *     cdef np.float64_t det, inv_det
 *     det = dot(e1, P)             # <<<<<<<<<<<<<<
 *     if(det > -DETERMINANT_EPS and det < DETERMINANT_EPS):
 *         return False
 */
  __pyx_v_det = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_e1, __pyx_v_P);

  /* "yt/utilities/lib/primitives.pyx":90
 *     cdef np.float64_t det, inv_det
 *     det = dot(e1, P)
 *     if(det > -DETERMINANT_EPS and det < DETERMINANT_EPS):             # <<<<<<<<<<<<<<
 *         return False
 *     inv_det = 1.0 / det
 */
  __pyx_t_2 = ((__pyx_v_det > (-__pyx_v_2yt_9utilities_3lib_10primitives_DETERMINANT_EPS)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_det < __pyx_v_2yt_9utilities_3lib_10primitives_DETERMINANT_EPS) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":91
 *     det = dot(e1, P)
 *     if(det > -DETERMINANT_EPS and det < DETERMINANT_EPS):
 *         return False             # <<<<<<<<<<<<<<
 *     inv_det = 1.0 / det
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":90
 *     cdef np.float64_t det, inv_det
 *     det = dot(e1, P)
 *     if(det > -DETERMINANT_EPS and det < DETERMINANT_EPS):             # <<<<<<<<<<<<<<
 *         return False
 *     inv_det = 1.0 / det
 */
  }

  /* "yt/utilities/lib/primitives.pyx":92
 *     if(det > -DETERMINANT_EPS and det < DETERMINANT_EPS):
 *         return False
 *     inv_det = 1.0 / det             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t T[3]
 */
  __pyx_v_inv_det = (1.0 / __pyx_v_det);

  /* "yt/utilities/lib/primitives.pyx":95
 * 
 *     cdef np.float64_t T[3]
 *     subtract(ray.origin, tri.p0, T)             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t u = dot(T, P) * inv_det
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_subtract(__pyx_v_ray->origin, __pyx_v_tri.p0, __pyx_v_T);

  /* "yt/utilities/lib/primitives.pyx":97
 *     subtract(ray.origin, tri.p0, T)
 * 
 *     cdef np.float64_t u = dot(T, P) * inv_det             # <<<<<<<<<<<<<<
 *     if(u < 0.0 or u > 1.0):
 *         return False
 */
  __pyx_v_u = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_T, __pyx_v_P) * __pyx_v_inv_det);

  /* "yt/utilities/lib/primitives.pyx":98
 * 
 *     cdef np.float64_t u = dot(T, P) * inv_det
 *     if(u < 0.0 or u > 1.0):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = ((__pyx_v_u < 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_u > 1.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":99
 *     cdef np.float64_t u = dot(T, P) * inv_det
 *     if(u < 0.0 or u > 1.0):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t Q[3]
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":98
 * 
 *     cdef np.float64_t u = dot(T, P) * inv_det
 *     if(u < 0.0 or u > 1.0):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "yt/utilities/lib/primitives.pyx":102
 * 
 *     cdef np.float64_t Q[3]
 *     cross(T, e1, Q)             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t v = dot(ray.direction, Q) * inv_det
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(__pyx_v_T, __pyx_v_e1, __pyx_v_Q);

  /* "yt/utilities/lib/primitives.pyx":104
 *     cross(T, e1, Q)
 * 
 *     cdef np.float64_t v = dot(ray.direction, Q) * inv_det             # <<<<<<<<<<<<<<
 *     if(v < 0.0 or u + v  > 1.0):
 *         return False
 */
  __pyx_v_v = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_ray->direction, __pyx_v_Q) * __pyx_v_inv_det);

  /* "yt/utilities/lib/primitives.pyx":105
 * 
 *     cdef np.float64_t v = dot(ray.direction, Q) * inv_det
 *     if(v < 0.0 or u + v  > 1.0):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = ((__pyx_v_v < 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_u + __pyx_v_v) > 1.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":106
 *     cdef np.float64_t v = dot(ray.direction, Q) * inv_det
 *     if(v < 0.0 or u + v  > 1.0):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t t = dot(e2, Q) * inv_det
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":105
 * 
 *     cdef np.float64_t v = dot(ray.direction, Q) * inv_det
 *     if(v < 0.0 or u + v  > 1.0):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "yt/utilities/lib/primitives.pyx":108
 *         return False
 * 
 *     cdef np.float64_t t = dot(e2, Q) * inv_det             # <<<<<<<<<<<<<<
 * 
 *     if(t > DETERMINANT_EPS and t < ray.t_far):
 */
  __pyx_v_t = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_e2, __pyx_v_Q) * __pyx_v_inv_det);

  /* "yt/utilities/lib/primitives.pyx":110
 *     cdef np.float64_t t = dot(e2, Q) * inv_det
 * 
 *     if(t > DETERMINANT_EPS and t < ray.t_far):             # <<<<<<<<<<<<<<
 *         ray.t_far = t
 *         ray.elem_id = tri.elem_id
 */
  __pyx_t_2 = ((__pyx_v_t > __pyx_v_2yt_9utilities_3lib_10primitives_DETERMINANT_EPS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_t < __pyx_v_ray->t_far) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":111
 * 
 *     if(t > DETERMINANT_EPS and t < ray.t_far):
 *         ray.t_far = t             # <<<<<<<<<<<<<<
 *         ray.elem_id = tri.elem_id
 *         return True
 */
    __pyx_v_ray->t_far = __pyx_v_t;

    /* "yt/utilities/lib/primitives.pyx":112
 *     if(t > DETERMINANT_EPS and t < ray.t_far):
 *         ray.t_far = t
 *         ray.elem_id = tri.elem_id             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_t_3 = __pyx_v_tri.elem_id;
    __pyx_v_ray->elem_id = __pyx_t_3;

    /* "yt/utilities/lib/primitives.pyx":113
 *         ray.t_far = t
 *         ray.elem_id = tri.elem_id
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     return False
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":110
 *     cdef np.float64_t t = dot(e2, Q) * inv_det
 * 
 *     if(t > DETERMINANT_EPS and t < ray.t_far):             # <<<<<<<<<<<<<<
 *         ray.t_far = t
 *         ray.elem_id = tri.elem_id
 */
  }

  /* "yt/utilities/lib/primitives.pyx":115
 *         return True
 * 
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":62
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_triangle_intersect(const void* primitives,             # <<<<<<<<<<<<<<
 *                                        const np.int64_t item,
 *                                        Ray* ray) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/primitives.pyx":121
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void triangle_centroid(const void *primitives,             # <<<<<<<<<<<<<<
 *                             const np.int64_t item,
 *                             np.float64_t[3] centroid) nogil:
 */

static void __pyx_f_2yt_9utilities_3lib_10primitives_triangle_centroid(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, __pyx_t_5numpy_float64_t *__pyx_v_centroid) {
  struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle __pyx_v_tri;
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":132
 *     '''
 * 
 *     cdef Triangle tri = (<Triangle*> primitives)[item]             # <<<<<<<<<<<<<<
 *     cdef np.int64_t i
 *     for i in range(3):
 */
  __pyx_v_tri = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":134
 *     cdef Triangle tri = (<Triangle*> primitives)[item]
 *     cdef np.int64_t i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         centroid[i] = (tri.p0[i] + tri.p1[i] + tri.p2[i]) / 3.0
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":135
 *     cdef np.int64_t i
 *     for i in range(3):
 *         centroid[i] = (tri.p0[i] + tri.p1[i] + tri.p2[i]) / 3.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_centroid[__pyx_v_i]) = ((((__pyx_v_tri.p0[__pyx_v_i]) + (__pyx_v_tri.p1[__pyx_v_i])) + (__pyx_v_tri.p2[__pyx_v_i])) / 3.0);
  }

  /* "yt/utilities/lib/primitives.pyx":121
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void triangle_centroid(const void *primitives,             # <<<<<<<<<<<<<<
 *                             const np.int64_t item,
 *                             np.float64_t[3] centroid) nogil:
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":141
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void triangle_bbox(const void *primitives,             # <<<<<<<<<<<<<<
 *                         const np.int64_t item,
 *                         BBox* bbox) nogil:
 */

static void __pyx_f_2yt_9utilities_3lib_10primitives_triangle_bbox(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox *__pyx_v_bbox) {
  struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle __pyx_v_tri;
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":152
 *     '''
 * 
 *     cdef Triangle tri = (<Triangle*> primitives)[item]             # <<<<<<<<<<<<<<
 *     cdef np.int64_t i
 *     for i in range(3):
 */
  __pyx_v_tri = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_Triangle *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":154
 *     cdef Triangle tri = (<Triangle*> primitives)[item]
 *     cdef np.int64_t i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         bbox.left_edge[i] = fmin(fmin(tri.p0[i], tri.p1[i]), tri.p2[i])
 *         bbox.right_edge[i] = fmax(fmax(tri.p0[i], tri.p1[i]), tri.p2[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":155
 *     cdef np.int64_t i
 *     for i in range(3):
 *         bbox.left_edge[i] = fmin(fmin(tri.p0[i], tri.p1[i]), tri.p2[i])             # <<<<<<<<<<<<<<
 *         bbox.right_edge[i] = fmax(fmax(tri.p0[i], tri.p1[i]), tri.p2[i])
 * 
 */
    (__pyx_v_bbox->left_edge[__pyx_v_i]) = fmin(fmin((__pyx_v_tri.p0[__pyx_v_i]), (__pyx_v_tri.p1[__pyx_v_i])), (__pyx_v_tri.p2[__pyx_v_i]));

    /* "yt/utilities/lib/primitives.pyx":156
 *     for i in range(3):
 *         bbox.left_edge[i] = fmin(fmin(tri.p0[i], tri.p1[i]), tri.p2[i])
 *         bbox.right_edge[i] = fmax(fmax(tri.p0[i], tri.p1[i]), tri.p2[i])             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_bbox->right_edge[__pyx_v_i]) = fmax(fmax((__pyx_v_tri.p0[__pyx_v_i]), (__pyx_v_tri.p1[__pyx_v_i])), (__pyx_v_tri.p2[__pyx_v_i]));
  }

  /* "yt/utilities/lib/primitives.pyx":141
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void triangle_bbox(const void *primitives,             # <<<<<<<<<<<<<<
 *                         const np.int64_t item,
 *                         BBox* bbox) nogil:
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":162
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceFunc(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                            const cython.floating u,
 *                            const cython.floating v,
 */

static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(float (*__pyx_v_verts)[3], float const __pyx_v_u, float const __pyx_v_v, float *__pyx_v_S) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":177
 *     '''
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         S[i] = 0.25*(1.0 - u)*(1.0 - v)*(-u - v - 1)*verts[0][i] + \
 *                0.25*(1.0 + u)*(1.0 - v)*( u - v - 1)*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":178
 *     cdef int i
 *     for i in range(3):
 *         S[i] = 0.25*(1.0 - u)*(1.0 - v)*(-u - v - 1)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *                0.25*(1.0 + u)*(1.0 - v)*( u - v - 1)*verts[1][i] + \
 *                0.25*(1.0 + u)*(1.0 + v)*( u + v - 1)*verts[2][i] + \
 */
    (__pyx_v_S[__pyx_v_i]) = (((((((((((0.25 * (1.0 - __pyx_v_u)) * (1.0 - __pyx_v_v)) * (((-__pyx_v_u) - __pyx_v_v) - 1.0)) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((((0.25 * (1.0 + __pyx_v_u)) * (1.0 - __pyx_v_v)) * ((__pyx_v_u - __pyx_v_v) - 1.0)) * ((__pyx_v_verts[1])[__pyx_v_i]))) + ((((0.25 * (1.0 + __pyx_v_u)) * (1.0 + __pyx_v_v)) * ((__pyx_v_u + __pyx_v_v) - 1.0)) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((0.25 * (1.0 - __pyx_v_u)) * (1.0 + __pyx_v_v)) * (((-__pyx_v_u) + __pyx_v_v) - 1.0)) * ((__pyx_v_verts[3])[__pyx_v_i]))) + (((0.5 * (1.0 - __pyx_v_u)) * (1.0 - (__pyx_v_v * __pyx_v_v))) * ((__pyx_v_verts[4])[__pyx_v_i]))) + (((0.5 * (1.0 - (__pyx_v_u * __pyx_v_u))) * (1.0 - __pyx_v_v)) * ((__pyx_v_verts[5])[__pyx_v_i]))) + (((0.5 * (1.0 + __pyx_v_u)) * (1.0 - (__pyx_v_v * __pyx_v_v))) * ((__pyx_v_verts[6])[__pyx_v_i]))) + (((0.5 * (1.0 - (__pyx_v_u * __pyx_v_u))) * (1.0 + __pyx_v_v)) * ((__pyx_v_verts[7])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":162
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceFunc(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                            const cython.floating u,
 *                            const cython.floating v,
 */

  /* function exit code */
}

static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(double (*__pyx_v_verts)[3], double const __pyx_v_u, double const __pyx_v_v, double *__pyx_v_S) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":177
 *     '''
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         S[i] = 0.25*(1.0 - u)*(1.0 - v)*(-u - v - 1)*verts[0][i] + \
 *                0.25*(1.0 + u)*(1.0 - v)*( u - v - 1)*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":178
 *     cdef int i
 *     for i in range(3):
 *         S[i] = 0.25*(1.0 - u)*(1.0 - v)*(-u - v - 1)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *                0.25*(1.0 + u)*(1.0 - v)*( u - v - 1)*verts[1][i] + \
 *                0.25*(1.0 + u)*(1.0 + v)*( u + v - 1)*verts[2][i] + \
 */
    (__pyx_v_S[__pyx_v_i]) = (((((((((((0.25 * (1.0 - __pyx_v_u)) * (1.0 - __pyx_v_v)) * (((-__pyx_v_u) - __pyx_v_v) - 1.0)) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((((0.25 * (1.0 + __pyx_v_u)) * (1.0 - __pyx_v_v)) * ((__pyx_v_u - __pyx_v_v) - 1.0)) * ((__pyx_v_verts[1])[__pyx_v_i]))) + ((((0.25 * (1.0 + __pyx_v_u)) * (1.0 + __pyx_v_v)) * ((__pyx_v_u + __pyx_v_v) - 1.0)) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((0.25 * (1.0 - __pyx_v_u)) * (1.0 + __pyx_v_v)) * (((-__pyx_v_u) + __pyx_v_v) - 1.0)) * ((__pyx_v_verts[3])[__pyx_v_i]))) + (((0.5 * (1.0 - __pyx_v_u)) * (1.0 - (__pyx_v_v * __pyx_v_v))) * ((__pyx_v_verts[4])[__pyx_v_i]))) + (((0.5 * (1.0 - (__pyx_v_u * __pyx_v_u))) * (1.0 - __pyx_v_v)) * ((__pyx_v_verts[5])[__pyx_v_i]))) + (((0.5 * (1.0 + __pyx_v_u)) * (1.0 - (__pyx_v_v * __pyx_v_v))) * ((__pyx_v_verts[6])[__pyx_v_i]))) + (((0.5 * (1.0 - (__pyx_v_u * __pyx_v_u))) * (1.0 + __pyx_v_v)) * ((__pyx_v_verts[7])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":162
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceFunc(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                            const cython.floating u,
 *                            const cython.floating v,
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":191
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceDerivU(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU(float (*__pyx_v_verts)[3], float const __pyx_v_u, float const __pyx_v_v, float *__pyx_v_Su) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":201
 *     '''
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Su[i] = (-0.25*(v - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \
 *                 (-0.25*(v - 1.0)*(u - v - 1) - 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":202
 *     cdef int i
 *     for i in range(3):
 *         Su[i] = (-0.25*(v - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \             # <<<<<<<<<<<<<<
 *                 (-0.25*(v - 1.0)*(u - v - 1) - 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 *                 ( 0.25*(v + 1.0)*(u + v - 1) + 0.25*(u + 1.0)*(v + 1.0))*verts[2][i] + \
 */
    (__pyx_v_Su[__pyx_v_i]) = (((((((((((-0.25 * (__pyx_v_v - 1.0)) * ((__pyx_v_u + __pyx_v_v) + 1.0)) - ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((((-0.25 * (__pyx_v_v - 1.0)) * ((__pyx_v_u - __pyx_v_v) - 1.0)) - ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[1])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_v + 1.0)) * ((__pyx_v_u + __pyx_v_v) - 1.0)) + ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_v + 1.0)) * ((__pyx_v_u - __pyx_v_v) + 1.0)) + ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((0.5 * ((__pyx_v_v * __pyx_v_v) - 1.0)) * ((__pyx_v_verts[4])[__pyx_v_i]))) + ((__pyx_v_u * (__pyx_v_v - 1.0)) * ((__pyx_v_verts[5])[__pyx_v_i]))) - ((0.5 * ((__pyx_v_v * __pyx_v_v) - 1.0)) * ((__pyx_v_verts[6])[__pyx_v_i]))) - ((__pyx_v_u * (__pyx_v_v + 1.0)) * ((__pyx_v_verts[7])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":191
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceDerivU(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU(double (*__pyx_v_verts)[3], double const __pyx_v_u, double const __pyx_v_v, double *__pyx_v_Su) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":201
 *     '''
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Su[i] = (-0.25*(v - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \
 *                 (-0.25*(v - 1.0)*(u - v - 1) - 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":202
 *     cdef int i
 *     for i in range(3):
 *         Su[i] = (-0.25*(v - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \             # <<<<<<<<<<<<<<
 *                 (-0.25*(v - 1.0)*(u - v - 1) - 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 *                 ( 0.25*(v + 1.0)*(u + v - 1) + 0.25*(u + 1.0)*(v + 1.0))*verts[2][i] + \
 */
    (__pyx_v_Su[__pyx_v_i]) = (((((((((((-0.25 * (__pyx_v_v - 1.0)) * ((__pyx_v_u + __pyx_v_v) + 1.0)) - ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((((-0.25 * (__pyx_v_v - 1.0)) * ((__pyx_v_u - __pyx_v_v) - 1.0)) - ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[1])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_v + 1.0)) * ((__pyx_v_u + __pyx_v_v) - 1.0)) + ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_v + 1.0)) * ((__pyx_v_u - __pyx_v_v) + 1.0)) + ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((0.5 * ((__pyx_v_v * __pyx_v_v) - 1.0)) * ((__pyx_v_verts[4])[__pyx_v_i]))) + ((__pyx_v_u * (__pyx_v_v - 1.0)) * ((__pyx_v_verts[5])[__pyx_v_i]))) - ((0.5 * ((__pyx_v_v * __pyx_v_v) - 1.0)) * ((__pyx_v_verts[6])[__pyx_v_i]))) - ((__pyx_v_u * (__pyx_v_v + 1.0)) * ((__pyx_v_verts[7])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":191
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceDerivU(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":213
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceDerivV(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV(float (*__pyx_v_verts)[3], float const __pyx_v_u, float const __pyx_v_v, float *__pyx_v_Sv) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":224
 * 
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Sv[i] = (-0.25*(u - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \
 *                 (-0.25*(u + 1.0)*(u - v - 1) + 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":225
 *     cdef int i
 *     for i in range(3):
 *         Sv[i] = (-0.25*(u - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \             # <<<<<<<<<<<<<<
 *                 (-0.25*(u + 1.0)*(u - v - 1) + 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 *                 ( 0.25*(u + 1.0)*(u + v - 1) + 0.25*(u + 1.0)*(v + 1.0))*verts[2][i] + \
 */
    (__pyx_v_Sv[__pyx_v_i]) = (((((((((((-0.25 * (__pyx_v_u - 1.0)) * ((__pyx_v_u + __pyx_v_v) + 1.0)) - ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((((-0.25 * (__pyx_v_u + 1.0)) * ((__pyx_v_u - __pyx_v_v) - 1.0)) + ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[1])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_u + 1.0)) * ((__pyx_v_u + __pyx_v_v) - 1.0)) + ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_u - 1.0)) * ((__pyx_v_u - __pyx_v_v) + 1.0)) - ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((0.5 * ((__pyx_v_u * __pyx_v_u) - 1.0)) * ((__pyx_v_verts[5])[__pyx_v_i]))) + ((__pyx_v_v * (__pyx_v_u - 1.0)) * ((__pyx_v_verts[4])[__pyx_v_i]))) - ((0.5 * ((__pyx_v_u * __pyx_v_u) - 1.0)) * ((__pyx_v_verts[7])[__pyx_v_i]))) - ((__pyx_v_v * (__pyx_v_u + 1.0)) * ((__pyx_v_verts[6])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":213
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceDerivV(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV(double (*__pyx_v_verts)[3], double const __pyx_v_u, double const __pyx_v_v, double *__pyx_v_Sv) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":224
 * 
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Sv[i] = (-0.25*(u - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \
 *                 (-0.25*(u + 1.0)*(u - v - 1) + 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":225
 *     cdef int i
 *     for i in range(3):
 *         Sv[i] = (-0.25*(u - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \             # <<<<<<<<<<<<<<
 *                 (-0.25*(u + 1.0)*(u - v - 1) + 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
 *                 ( 0.25*(u + 1.0)*(u + v - 1) + 0.25*(u + 1.0)*(v + 1.0))*verts[2][i] + \
 */
    (__pyx_v_Sv[__pyx_v_i]) = (((((((((((-0.25 * (__pyx_v_u - 1.0)) * ((__pyx_v_u + __pyx_v_v) + 1.0)) - ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((((-0.25 * (__pyx_v_u + 1.0)) * ((__pyx_v_u - __pyx_v_v) - 1.0)) + ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v - 1.0))) * ((__pyx_v_verts[1])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_u + 1.0)) * ((__pyx_v_u + __pyx_v_v) - 1.0)) + ((0.25 * (__pyx_v_u + 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((0.25 * (__pyx_v_u - 1.0)) * ((__pyx_v_u - __pyx_v_v) + 1.0)) - ((0.25 * (__pyx_v_u - 1.0)) * (__pyx_v_v + 1.0))) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((0.5 * ((__pyx_v_u * __pyx_v_u) - 1.0)) * ((__pyx_v_verts[5])[__pyx_v_i]))) + ((__pyx_v_v * (__pyx_v_u - 1.0)) * ((__pyx_v_verts[4])[__pyx_v_i]))) - ((0.5 * ((__pyx_v_u * __pyx_v_u) - 1.0)) * ((__pyx_v_verts[7])[__pyx_v_i]))) - ((__pyx_v_v * (__pyx_v_u + 1.0)) * ((__pyx_v_verts[6])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":213
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patchSurfaceDerivV(const cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":236
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef RayHitData compute_patch_hit(cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                                   cython.floating[3] ray_origin,
 *                                   cython.floating[3] ray_direction) nogil:
 */

static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit(float (*__pyx_v_verts)[3], float *__pyx_v_ray_origin, float *__pyx_v_ray_direction) {
  float __pyx_v_n[3];
  float __pyx_v_N1[3];
  float __pyx_v_N2[3];
  float __pyx_v_A;
  long __pyx_v_i;
  float __pyx_v_d1;
  float __pyx_v_d2;
  float __pyx_v_u;
  float __pyx_v_v;
  float __pyx_v_S[3];
  float __pyx_v_fu;
  float __pyx_v_fv;
  float __pyx_v_err;
  float __pyx_v_tol;
  int __pyx_v_iterations;
  int __pyx_v_max_iter;
  float __pyx_v_Su[3];
  float __pyx_v_Sv[3];
  float __pyx_v_J11;
  float __pyx_v_J12;
  float __pyx_v_J21;
  float __pyx_v_J22;
  float __pyx_v_det;
  float __pyx_v_t;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_v_hd;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_r;
  long __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":248
 *     # first we compute the two planes that define the ray.
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)             # <<<<<<<<<<<<<<
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A
 */
  __pyx_v_A = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_ray_direction, __pyx_v_ray_direction);

  /* "yt/utilities/lib/primitives.pyx":249
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         n[i] = ray_direction[i] / A
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":250
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A             # <<<<<<<<<<<<<<
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 */
    (__pyx_v_n[__pyx_v_i]) = ((__pyx_v_ray_direction[__pyx_v_i]) / __pyx_v_A);
  }

  /* "yt/utilities/lib/primitives.pyx":252
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[1]))) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[2]))) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/utilities/lib/primitives.pyx":253
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]             # <<<<<<<<<<<<<<
 *         N1[1] =-n[0]
 *         N1[2] = 0.0
 */
    (__pyx_v_N1[0]) = (__pyx_v_n[1]);

    /* "yt/utilities/lib/primitives.pyx":254
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]
 *         N1[1] =-n[0]             # <<<<<<<<<<<<<<
 *         N1[2] = 0.0
 *     else:
 */
    (__pyx_v_N1[1]) = (-(__pyx_v_n[0]));

    /* "yt/utilities/lib/primitives.pyx":255
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 *         N1[2] = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         N1[0] = 0.0
 */
    (__pyx_v_N1[2]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":252
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
    goto __pyx_L5;
  }

  /* "yt/utilities/lib/primitives.pyx":257
 *         N1[2] = 0.0
 *     else:
 *         N1[0] = 0.0             # <<<<<<<<<<<<<<
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 */
  /*else*/ {
    (__pyx_v_N1[0]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":258
 *     else:
 *         N1[0] = 0.0
 *         N1[1] = n[2]             # <<<<<<<<<<<<<<
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)
 */
    (__pyx_v_N1[1]) = (__pyx_v_n[2]);

    /* "yt/utilities/lib/primitives.pyx":259
 *         N1[0] = 0.0
 *         N1[1] = n[2]
 *         N1[2] =-n[1]             # <<<<<<<<<<<<<<
 *     cross(N1, n, N2)
 * 
 */
    (__pyx_v_N1[2]) = (-(__pyx_v_n[1]));
  }
  __pyx_L5:;

  /* "yt/utilities/lib/primitives.pyx":260
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)             # <<<<<<<<<<<<<<
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 */
  __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(__pyx_v_N1, __pyx_v_n, __pyx_v_N2);

  /* "yt/utilities/lib/primitives.pyx":262
 *     cross(N1, n, N2)
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)             # <<<<<<<<<<<<<<
 *     cdef cython.floating d2 = -dot(N2, ray_origin)
 * 
 */
  __pyx_v_d1 = (-__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":263
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 *     cdef cython.floating d2 = -dot(N2, ray_origin)             # <<<<<<<<<<<<<<
 * 
 *     # the initial guess is set to zero
 */
  __pyx_v_d2 = (-__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":266
 * 
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 */
  __pyx_v_u = 0.0;

  /* "yt/utilities/lib/primitives.pyx":267
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0
 *     cdef cython.floating v = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] S
 *     patchSurfaceFunc(verts, u, v, S)
 */
  __pyx_v_v = 0.0;

  /* "yt/utilities/lib/primitives.pyx":269
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 *     patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 */
  __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

  /* "yt/utilities/lib/primitives.pyx":270
 *     cdef cython.floating[3] S
 *     patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 */
  __pyx_v_fu = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

  /* "yt/utilities/lib/primitives.pyx":271
 *     patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 * 
 */
  __pyx_v_fv = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

  /* "yt/utilities/lib/primitives.pyx":272
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 * 
 *     # begin Newton iteration
 */
  __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

  /* "yt/utilities/lib/primitives.pyx":275
 * 
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5             # <<<<<<<<<<<<<<
 *     cdef int iterations = 0
 *     cdef int max_iter = 10
 */
  __pyx_v_tol = 1.0e-5;

  /* "yt/utilities/lib/primitives.pyx":276
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0             # <<<<<<<<<<<<<<
 *     cdef int max_iter = 10
 *     cdef cython.floating[3] Su
 */
  __pyx_v_iterations = 0;

  /* "yt/utilities/lib/primitives.pyx":277
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0
 *     cdef int max_iter = 10             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] Su
 *     cdef cython.floating[3] Sv
 */
  __pyx_v_max_iter = 10;

  /* "yt/utilities/lib/primitives.pyx":281
 *     cdef cython.floating[3] Sv
 *     cdef cython.floating J11, J12, J21, J22, det
 *     while ((err > tol) and (iterations < max_iter)):             # <<<<<<<<<<<<<<
 *         # compute the Jacobian
 *         patchSurfaceDerivU(verts, u, v, Su)
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_err > __pyx_v_tol) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_3 = ((__pyx_v_iterations < __pyx_v_max_iter) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L10_bool_binop_done:;
    if (!__pyx_t_2) break;

    /* "yt/utilities/lib/primitives.pyx":283
 *     while ((err > tol) and (iterations < max_iter)):
 *         # compute the Jacobian
 *         patchSurfaceDerivU(verts, u, v, Su)             # <<<<<<<<<<<<<<
 *         patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 */
    __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":284
 *         # compute the Jacobian
 *         patchSurfaceDerivU(verts, u, v, Su)
 *         patchSurfaceDerivV(verts, u, v, Sv)             # <<<<<<<<<<<<<<
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 */
    __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":285
 *         patchSurfaceDerivU(verts, u, v, Su)
 *         patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)             # <<<<<<<<<<<<<<
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 */
    __pyx_v_J11 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":286
 *         patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)             # <<<<<<<<<<<<<<
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 */
    __pyx_v_J12 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":287
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)             # <<<<<<<<<<<<<<
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)
 */
    __pyx_v_J21 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":288
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)             # <<<<<<<<<<<<<<
 *         det = (J11*J22 - J12*J21)
 * 
 */
    __pyx_v_J22 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":289
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)             # <<<<<<<<<<<<<<
 * 
 *         # update the u, v values
 */
    __pyx_v_det = ((__pyx_v_J11 * __pyx_v_J22) - (__pyx_v_J12 * __pyx_v_J21));

    /* "yt/utilities/lib/primitives.pyx":292
 * 
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det             # <<<<<<<<<<<<<<
 *         v -= (-J21*fu + J11*fv) / det
 * 
 */
    __pyx_v_u = (__pyx_v_u - (((__pyx_v_J22 * __pyx_v_fu) - (__pyx_v_J12 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":293
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det
 *         v -= (-J21*fu + J11*fv) / det             # <<<<<<<<<<<<<<
 * 
 *         patchSurfaceFunc(verts, u, v, S)
 */
    __pyx_v_v = (__pyx_v_v - ((((-__pyx_v_J21) * __pyx_v_fu) + (__pyx_v_J11 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":295
 *         v -= (-J21*fu + J11*fv) / det
 * 
 *         patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2
 */
    __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

    /* "yt/utilities/lib/primitives.pyx":296
 * 
 *         patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *         fv = dot(N2, S) + d2
 * 
 */
    __pyx_v_fu = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

    /* "yt/utilities/lib/primitives.pyx":297
 *         patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 */
    __pyx_v_fv = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

    /* "yt/utilities/lib/primitives.pyx":299
 *         fv = dot(N2, S) + d2
 * 
 *         err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 *         iterations += 1
 * 
 */
    __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

    /* "yt/utilities/lib/primitives.pyx":300
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 *         iterations += 1             # <<<<<<<<<<<<<<
 * 
 *     # t is the distance along the ray to this hit
 */
    __pyx_v_iterations = (__pyx_v_iterations + 1);
  }

  /* "yt/utilities/lib/primitives.pyx":303
 * 
 *     # t is the distance along the ray to this hit
 *     cdef cython.floating t = distance(S, ray_origin) / L2_norm(ray_direction)             # <<<<<<<<<<<<<<
 * 
 *     # return hit data
 */
  __pyx_v_t = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(__pyx_v_S, __pyx_v_ray_origin) / __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(__pyx_v_ray_direction));

  /* "yt/utilities/lib/primitives.pyx":307
 *     # return hit data
 *     cdef RayHitData hd
 *     hd.u = u             # <<<<<<<<<<<<<<
 *     hd.v = v
 *     hd.t = t
 */
  __pyx_v_hd.u = __pyx_v_u;

  /* "yt/utilities/lib/primitives.pyx":308
 *     cdef RayHitData hd
 *     hd.u = u
 *     hd.v = v             # <<<<<<<<<<<<<<
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 */
  __pyx_v_hd.v = __pyx_v_v;

  /* "yt/utilities/lib/primitives.pyx":309
 *     hd.u = u
 *     hd.v = v
 *     hd.t = t             # <<<<<<<<<<<<<<
 *     hd.converged = (iterations < max_iter)
 *     return hd
 */
  __pyx_v_hd.t = __pyx_v_t;

  /* "yt/utilities/lib/primitives.pyx":310
 *     hd.v = v
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)             # <<<<<<<<<<<<<<
 *     return hd
 * 
 */
  __pyx_v_hd.converged = (__pyx_v_iterations < __pyx_v_max_iter);

  /* "yt/utilities/lib/primitives.pyx":311
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 *     return hd             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_hd;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":236
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef RayHitData compute_patch_hit(cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                                   cython.floating[3] ray_origin,
 *                                   cython.floating[3] ray_direction) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit(double (*__pyx_v_verts)[3], double *__pyx_v_ray_origin, double *__pyx_v_ray_direction) {
  double __pyx_v_n[3];
  double __pyx_v_N1[3];
  double __pyx_v_N2[3];
  double __pyx_v_A;
  long __pyx_v_i;
  double __pyx_v_d1;
  double __pyx_v_d2;
  double __pyx_v_u;
  double __pyx_v_v;
  double __pyx_v_S[3];
  double __pyx_v_fu;
  double __pyx_v_fv;
  double __pyx_v_err;
  double __pyx_v_tol;
  int __pyx_v_iterations;
  int __pyx_v_max_iter;
  double __pyx_v_Su[3];
  double __pyx_v_Sv[3];
  double __pyx_v_J11;
  double __pyx_v_J12;
  double __pyx_v_J21;
  double __pyx_v_J22;
  double __pyx_v_det;
  double __pyx_v_t;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_v_hd;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_r;
  long __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":248
 *     # first we compute the two planes that define the ray.
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)             # <<<<<<<<<<<<<<
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A
 */
  __pyx_v_A = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_ray_direction, __pyx_v_ray_direction);

  /* "yt/utilities/lib/primitives.pyx":249
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         n[i] = ray_direction[i] / A
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":250
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A             # <<<<<<<<<<<<<<
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 */
    (__pyx_v_n[__pyx_v_i]) = ((__pyx_v_ray_direction[__pyx_v_i]) / __pyx_v_A);
  }

  /* "yt/utilities/lib/primitives.pyx":252
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[1]))) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[2]))) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/utilities/lib/primitives.pyx":253
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]             # <<<<<<<<<<<<<<
 *         N1[1] =-n[0]
 *         N1[2] = 0.0
 */
    (__pyx_v_N1[0]) = (__pyx_v_n[1]);

    /* "yt/utilities/lib/primitives.pyx":254
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]
 *         N1[1] =-n[0]             # <<<<<<<<<<<<<<
 *         N1[2] = 0.0
 *     else:
 */
    (__pyx_v_N1[1]) = (-(__pyx_v_n[0]));

    /* "yt/utilities/lib/primitives.pyx":255
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 *         N1[2] = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         N1[0] = 0.0
 */
    (__pyx_v_N1[2]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":252
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
    goto __pyx_L5;
  }

  /* "yt/utilities/lib/primitives.pyx":257
 *         N1[2] = 0.0
 *     else:
 *         N1[0] = 0.0             # <<<<<<<<<<<<<<
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 */
  /*else*/ {
    (__pyx_v_N1[0]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":258
 *     else:
 *         N1[0] = 0.0
 *         N1[1] = n[2]             # <<<<<<<<<<<<<<
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)
 */
    (__pyx_v_N1[1]) = (__pyx_v_n[2]);

    /* "yt/utilities/lib/primitives.pyx":259
 *         N1[0] = 0.0
 *         N1[1] = n[2]
 *         N1[2] =-n[1]             # <<<<<<<<<<<<<<
 *     cross(N1, n, N2)
 * 
 */
    (__pyx_v_N1[2]) = (-(__pyx_v_n[1]));
  }
  __pyx_L5:;

  /* "yt/utilities/lib/primitives.pyx":260
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)             # <<<<<<<<<<<<<<
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(__pyx_v_N1, __pyx_v_n, __pyx_v_N2);

  /* "yt/utilities/lib/primitives.pyx":262
 *     cross(N1, n, N2)
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)             # <<<<<<<<<<<<<<
 *     cdef cython.floating d2 = -dot(N2, ray_origin)
 * 
 */
  __pyx_v_d1 = (-__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":263
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 *     cdef cython.floating d2 = -dot(N2, ray_origin)             # <<<<<<<<<<<<<<
 * 
 *     # the initial guess is set to zero
 */
  __pyx_v_d2 = (-__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":266
 * 
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 */
  __pyx_v_u = 0.0;

  /* "yt/utilities/lib/primitives.pyx":267
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0
 *     cdef cython.floating v = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] S
 *     patchSurfaceFunc(verts, u, v, S)
 */
  __pyx_v_v = 0.0;

  /* "yt/utilities/lib/primitives.pyx":269
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 *     patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

  /* "yt/utilities/lib/primitives.pyx":270
 *     cdef cython.floating[3] S
 *     patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 */
  __pyx_v_fu = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

  /* "yt/utilities/lib/primitives.pyx":271
 *     patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 * 
 */
  __pyx_v_fv = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

  /* "yt/utilities/lib/primitives.pyx":272
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 * 
 *     # begin Newton iteration
 */
  __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

  /* "yt/utilities/lib/primitives.pyx":275
 * 
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5             # <<<<<<<<<<<<<<
 *     cdef int iterations = 0
 *     cdef int max_iter = 10
 */
  __pyx_v_tol = 1.0e-5;

  /* "yt/utilities/lib/primitives.pyx":276
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0             # <<<<<<<<<<<<<<
 *     cdef int max_iter = 10
 *     cdef cython.floating[3] Su
 */
  __pyx_v_iterations = 0;

  /* "yt/utilities/lib/primitives.pyx":277
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0
 *     cdef int max_iter = 10             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] Su
 *     cdef cython.floating[3] Sv
 */
  __pyx_v_max_iter = 10;

  /* "yt/utilities/lib/primitives.pyx":281
 *     cdef cython.floating[3] Sv
 *     cdef cython.floating J11, J12, J21, J22, det
 *     while ((err > tol) and (iterations < max_iter)):             # <<<<<<<<<<<<<<
 *         # compute the Jacobian
 *         patchSurfaceDerivU(verts, u, v, Su)
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_err > __pyx_v_tol) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_3 = ((__pyx_v_iterations < __pyx_v_max_iter) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L10_bool_binop_done:;
    if (!__pyx_t_2) break;

    /* "yt/utilities/lib/primitives.pyx":283
 *     while ((err > tol) and (iterations < max_iter)):
 *         # compute the Jacobian
 *         patchSurfaceDerivU(verts, u, v, Su)             # <<<<<<<<<<<<<<
 *         patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 */
    __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":284
 *         # compute the Jacobian
 *         patchSurfaceDerivU(verts, u, v, Su)
 *         patchSurfaceDerivV(verts, u, v, Sv)             # <<<<<<<<<<<<<<
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 */
    __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":285
 *         patchSurfaceDerivU(verts, u, v, Su)
 *         patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)             # <<<<<<<<<<<<<<
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 */
    __pyx_v_J11 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":286
 *         patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)             # <<<<<<<<<<<<<<
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 */
    __pyx_v_J12 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":287
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)             # <<<<<<<<<<<<<<
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)
 */
    __pyx_v_J21 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":288
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)             # <<<<<<<<<<<<<<
 *         det = (J11*J22 - J12*J21)
 * 
 */
    __pyx_v_J22 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":289
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)             # <<<<<<<<<<<<<<
 * 
 *         # update the u, v values
 */
    __pyx_v_det = ((__pyx_v_J11 * __pyx_v_J22) - (__pyx_v_J12 * __pyx_v_J21));

    /* "yt/utilities/lib/primitives.pyx":292
 * 
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det             # <<<<<<<<<<<<<<
 *         v -= (-J21*fu + J11*fv) / det
 * 
 */
    __pyx_v_u = (__pyx_v_u - (((__pyx_v_J22 * __pyx_v_fu) - (__pyx_v_J12 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":293
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det
 *         v -= (-J21*fu + J11*fv) / det             # <<<<<<<<<<<<<<
 * 
 *         patchSurfaceFunc(verts, u, v, S)
 */
    __pyx_v_v = (__pyx_v_v - ((((-__pyx_v_J21) * __pyx_v_fu) + (__pyx_v_J11 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":295
 *         v -= (-J21*fu + J11*fv) / det
 * 
 *         patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2
 */
    __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

    /* "yt/utilities/lib/primitives.pyx":296
 * 
 *         patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *         fv = dot(N2, S) + d2
 * 
 */
    __pyx_v_fu = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

    /* "yt/utilities/lib/primitives.pyx":297
 *         patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 */
    __pyx_v_fv = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

    /* "yt/utilities/lib/primitives.pyx":299
 *         fv = dot(N2, S) + d2
 * 
 *         err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 *         iterations += 1
 * 
 */
    __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

    /* "yt/utilities/lib/primitives.pyx":300
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 *         iterations += 1             # <<<<<<<<<<<<<<
 * 
 *     # t is the distance along the ray to this hit
 */
    __pyx_v_iterations = (__pyx_v_iterations + 1);
  }

  /* "yt/utilities/lib/primitives.pyx":303
 * 
 *     # t is the distance along the ray to this hit
 *     cdef cython.floating t = distance(S, ray_origin) / L2_norm(ray_direction)             # <<<<<<<<<<<<<<
 * 
 *     # return hit data
 */
  __pyx_v_t = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(__pyx_v_S, __pyx_v_ray_origin) / __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(__pyx_v_ray_direction));

  /* "yt/utilities/lib/primitives.pyx":307
 *     # return hit data
 *     cdef RayHitData hd
 *     hd.u = u             # <<<<<<<<<<<<<<
 *     hd.v = v
 *     hd.t = t
 */
  __pyx_v_hd.u = __pyx_v_u;

  /* "yt/utilities/lib/primitives.pyx":308
 *     cdef RayHitData hd
 *     hd.u = u
 *     hd.v = v             # <<<<<<<<<<<<<<
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 */
  __pyx_v_hd.v = __pyx_v_v;

  /* "yt/utilities/lib/primitives.pyx":309
 *     hd.u = u
 *     hd.v = v
 *     hd.t = t             # <<<<<<<<<<<<<<
 *     hd.converged = (iterations < max_iter)
 *     return hd
 */
  __pyx_v_hd.t = __pyx_v_t;

  /* "yt/utilities/lib/primitives.pyx":310
 *     hd.v = v
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)             # <<<<<<<<<<<<<<
 *     return hd
 * 
 */
  __pyx_v_hd.converged = (__pyx_v_iterations < __pyx_v_max_iter);

  /* "yt/utilities/lib/primitives.pyx":311
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 *     return hd             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_hd;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":236
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef RayHitData compute_patch_hit(cython.floating[8][3] verts,             # <<<<<<<<<<<<<<
 *                                   cython.floating[3] ray_origin,
 *                                   cython.floating[3] ray_direction) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/primitives.pyx":317
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_patch_intersect(const void* primitives,             # <<<<<<<<<<<<<<
 *                                     const np.int64_t item,
 *                                     Ray* ray) nogil:
 */

static __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_10primitives_ray_patch_intersect(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *__pyx_v_ray) {
  struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch __pyx_v_patch;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_v_hd;
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5numpy_float64_t __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;

  /* "yt/utilities/lib/primitives.pyx":329
 * 
 *     '''
 *     cdef Patch patch = (<Patch*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     cdef RayHitData hd = compute_patch_hit(patch.v, ray.origin, ray.direction)
 */
  __pyx_v_patch = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":331
 *     cdef Patch patch = (<Patch*> primitives)[item]
 * 
 *     cdef RayHitData hd = compute_patch_hit(patch.v, ray.origin, ray.direction)             # <<<<<<<<<<<<<<
 * 
 *     # only count this is it's the closest hit
 */
  __pyx_v_hd = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit(__pyx_v_patch.v, __pyx_v_ray->origin, __pyx_v_ray->direction);

  /* "yt/utilities/lib/primitives.pyx":334
 * 
 *     # only count this is it's the closest hit
 *     if (hd.t < ray.t_near or hd.t > ray.t_far):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = ((__pyx_v_hd.t < __pyx_v_ray->t_near) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_hd.t > __pyx_v_ray->t_far) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":335
 *     # only count this is it's the closest hit
 *     if (hd.t < ray.t_near or hd.t > ray.t_far):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     if (fabs(hd.u) <= 1.0 and fabs(hd.v) <= 1.0 and hd.converged):
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":334
 * 
 *     # only count this is it's the closest hit
 *     if (hd.t < ray.t_near or hd.t > ray.t_far):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "yt/utilities/lib/primitives.pyx":337
 *         return False
 * 
 *     if (fabs(hd.u) <= 1.0 and fabs(hd.v) <= 1.0 and hd.converged):             # <<<<<<<<<<<<<<
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t
 */
  __pyx_t_2 = ((fabs(__pyx_v_hd.u) <= 1.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((fabs(__pyx_v_hd.v) <= 1.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_hd.converged != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":339
 *     if (fabs(hd.u) <= 1.0 and fabs(hd.v) <= 1.0 and hd.converged):
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t             # <<<<<<<<<<<<<<
 *         ray.elem_id = patch.elem_id
 *         return True
 */
    __pyx_t_3 = __pyx_v_hd.t;
    __pyx_v_ray->t_far = __pyx_t_3;

    /* "yt/utilities/lib/primitives.pyx":340
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t
 *         ray.elem_id = patch.elem_id             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_t_4 = __pyx_v_patch.elem_id;
    __pyx_v_ray->elem_id = __pyx_t_4;

    /* "yt/utilities/lib/primitives.pyx":341
 *         ray.t_far = hd.t
 *         ray.elem_id = patch.elem_id
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     return False
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":337
 *         return False
 * 
 *     if (fabs(hd.u) <= 1.0 and fabs(hd.v) <= 1.0 and hd.converged):             # <<<<<<<<<<<<<<
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t
 */
  }

  /* "yt/utilities/lib/primitives.pyx":343
 *         return True
 * 
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":317
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_patch_intersect(const void* primitives,             # <<<<<<<<<<<<<<
 *                                     const np.int64_t item,
 *                                     Ray* ray) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/primitives.pyx":349
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patch_centroid(const void *primitives,             # <<<<<<<<<<<<<<
 *                          const np.int64_t item,
 *                          np.float64_t[3] centroid) nogil:
 */

static void __pyx_f_2yt_9utilities_3lib_10primitives_patch_centroid(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, __pyx_t_5numpy_float64_t *__pyx_v_centroid) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_j;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch __pyx_v_patch;
  __pyx_t_5numpy_int64_t __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":361
 * 
 *     cdef np.int64_t i, j
 *     cdef Patch patch = (<Patch*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(3):
 */
  __pyx_v_patch = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":363
 *     cdef Patch patch = (<Patch*> primitives)[item]
 * 
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         centroid[j] = 0.0
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":364
 * 
 *     for j in range(3):
 *         centroid[j] = 0.0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(8):
 */
    (__pyx_v_centroid[__pyx_v_j]) = 0.0;
  }

  /* "yt/utilities/lib/primitives.pyx":366
 *         centroid[j] = 0.0
 * 
 *     for i in range(8):             # <<<<<<<<<<<<<<
 *         for j in range(3):
 *             centroid[j] += patch.v[i][j]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 8; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":367
 * 
 *     for i in range(8):
 *         for j in range(3):             # <<<<<<<<<<<<<<
 *             centroid[j] += patch.v[i][j]
 * 
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/utilities/lib/primitives.pyx":368
 *     for i in range(8):
 *         for j in range(3):
 *             centroid[j] += patch.v[i][j]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(3):
 */
      __pyx_t_3 = __pyx_v_j;
      (__pyx_v_centroid[__pyx_t_3]) = ((__pyx_v_centroid[__pyx_t_3]) + ((__pyx_v_patch.v[__pyx_v_i])[__pyx_v_j]));
    }
  }

  /* "yt/utilities/lib/primitives.pyx":370
 *             centroid[j] += patch.v[i][j]
 * 
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         centroid[j] /= 8.0
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":371
 * 
 *     for j in range(3):
 *         centroid[j] /= 8.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_v_j;
    (__pyx_v_centroid[__pyx_t_2]) = ((__pyx_v_centroid[__pyx_t_2]) / 8.0);
  }

  /* "yt/utilities/lib/primitives.pyx":349
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patch_centroid(const void *primitives,             # <<<<<<<<<<<<<<
 *                          const np.int64_t item,
 *                          np.float64_t[3] centroid) nogil:
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":377
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patch_bbox(const void *primitives,             # <<<<<<<<<<<<<<
 *                     const np.int64_t item,
 *                      BBox* bbox) nogil:
 */

static void __pyx_f_2yt_9utilities_3lib_10primitives_patch_bbox(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox *__pyx_v_bbox) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_j;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch __pyx_v_patch;
  __pyx_t_5numpy_int64_t __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;

  /* "yt/utilities/lib/primitives.pyx":390
 * 
 *     cdef np.int64_t i, j
 *     cdef Patch patch = (<Patch*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(3):
 */
  __pyx_v_patch = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_Patch *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":392
 *     cdef Patch patch = (<Patch*> primitives)[item]
 * 
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         bbox.left_edge[j] = patch.v[0][j]
 *         bbox.right_edge[j] = patch.v[0][j]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":393
 * 
 *     for j in range(3):
 *         bbox.left_edge[j] = patch.v[0][j]             # <<<<<<<<<<<<<<
 *         bbox.right_edge[j] = patch.v[0][j]
 * 
 */
    (__pyx_v_bbox->left_edge[__pyx_v_j]) = ((__pyx_v_patch.v[0])[__pyx_v_j]);

    /* "yt/utilities/lib/primitives.pyx":394
 *     for j in range(3):
 *         bbox.left_edge[j] = patch.v[0][j]
 *         bbox.right_edge[j] = patch.v[0][j]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(1, 8):
 */
    (__pyx_v_bbox->right_edge[__pyx_v_j]) = ((__pyx_v_patch.v[0])[__pyx_v_j]);
  }

  /* "yt/utilities/lib/primitives.pyx":396
 *         bbox.right_edge[j] = patch.v[0][j]
 * 
 *     for i in range(1, 8):             # <<<<<<<<<<<<<<
 *         for j in range(3):
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], patch.v[i][j])
 */
  for (__pyx_t_1 = 1; __pyx_t_1 < 8; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":397
 * 
 *     for i in range(1, 8):
 *         for j in range(3):             # <<<<<<<<<<<<<<
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], patch.v[i][j])
 *             bbox.right_edge[j] = fmax(bbox.right_edge[j], patch.v[i][j])
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/utilities/lib/primitives.pyx":398
 *     for i in range(1, 8):
 *         for j in range(3):
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], patch.v[i][j])             # <<<<<<<<<<<<<<
 *             bbox.right_edge[j] = fmax(bbox.right_edge[j], patch.v[i][j])
 * 
 */
      (__pyx_v_bbox->left_edge[__pyx_v_j]) = fmin((__pyx_v_bbox->left_edge[__pyx_v_j]), ((__pyx_v_patch.v[__pyx_v_i])[__pyx_v_j]));

      /* "yt/utilities/lib/primitives.pyx":399
 *         for j in range(3):
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], patch.v[i][j])
 *             bbox.right_edge[j] = fmax(bbox.right_edge[j], patch.v[i][j])             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_bbox->right_edge[__pyx_v_j]) = fmax((__pyx_v_bbox->right_edge[__pyx_v_j]), ((__pyx_v_patch.v[__pyx_v_i])[__pyx_v_j]));
    }
  }

  /* "yt/utilities/lib/primitives.pyx":377
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void patch_bbox(const void *primitives,             # <<<<<<<<<<<<<<
 *                     const np.int64_t item,
 *                      BBox* bbox) nogil:
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":405
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceFunc(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                            const cython.floating u,
 *                            const cython.floating v,
 */

static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(float (*__pyx_v_verts)[3], float const __pyx_v_u, float const __pyx_v_v, float *__pyx_v_S) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":412
 *     cdef int i
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         S[i] = (1.0 - 3.0*u + 2.0*u*u - 3.0*v + 2.0*v*v + 4.0*u*v)*verts[0][i] + \
 *              (-u + 2.0*u*u)*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":413
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):
 *         S[i] = (1.0 - 3.0*u + 2.0*u*u - 3.0*v + 2.0*v*v + 4.0*u*v)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *              (-u + 2.0*u*u)*verts[1][i] + \
 *              (-v + 2.0*v*v)*verts[2][i] + \
 */
    (__pyx_v_S[__pyx_v_i]) = (((((((((((1.0 - (3.0 * __pyx_v_u)) + ((2.0 * __pyx_v_u) * __pyx_v_u)) - (3.0 * __pyx_v_v)) + ((2.0 * __pyx_v_v) * __pyx_v_v)) + ((4.0 * __pyx_v_u) * __pyx_v_v)) * ((__pyx_v_verts[0])[__pyx_v_i])) + (((-__pyx_v_u) + ((2.0 * __pyx_v_u) * __pyx_v_u)) * ((__pyx_v_verts[1])[__pyx_v_i]))) + (((-__pyx_v_v) + ((2.0 * __pyx_v_v) * __pyx_v_v)) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((4.0 * __pyx_v_u) - ((4.0 * __pyx_v_u) * __pyx_v_u)) - ((4.0 * __pyx_v_u) * __pyx_v_v)) * ((__pyx_v_verts[3])[__pyx_v_i]))) + (((4.0 * __pyx_v_u) * __pyx_v_v) * ((__pyx_v_verts[4])[__pyx_v_i]))) + ((((4.0 * __pyx_v_v) - ((4.0 * __pyx_v_v) * __pyx_v_v)) - ((4.0 * __pyx_v_u) * __pyx_v_v)) * ((__pyx_v_verts[5])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":405
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceFunc(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                            const cython.floating u,
 *                            const cython.floating v,
 */

  /* function exit code */
}

static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(double (*__pyx_v_verts)[3], double const __pyx_v_u, double const __pyx_v_v, double *__pyx_v_S) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":412
 *     cdef int i
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         S[i] = (1.0 - 3.0*u + 2.0*u*u - 3.0*v + 2.0*v*v + 4.0*u*v)*verts[0][i] + \
 *              (-u + 2.0*u*u)*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":413
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):
 *         S[i] = (1.0 - 3.0*u + 2.0*u*u - 3.0*v + 2.0*v*v + 4.0*u*v)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *              (-u + 2.0*u*u)*verts[1][i] + \
 *              (-v + 2.0*v*v)*verts[2][i] + \
 */
    (__pyx_v_S[__pyx_v_i]) = (((((((((((1.0 - (3.0 * __pyx_v_u)) + ((2.0 * __pyx_v_u) * __pyx_v_u)) - (3.0 * __pyx_v_v)) + ((2.0 * __pyx_v_v) * __pyx_v_v)) + ((4.0 * __pyx_v_u) * __pyx_v_v)) * ((__pyx_v_verts[0])[__pyx_v_i])) + (((-__pyx_v_u) + ((2.0 * __pyx_v_u) * __pyx_v_u)) * ((__pyx_v_verts[1])[__pyx_v_i]))) + (((-__pyx_v_v) + ((2.0 * __pyx_v_v) * __pyx_v_v)) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((((4.0 * __pyx_v_u) - ((4.0 * __pyx_v_u) * __pyx_v_u)) - ((4.0 * __pyx_v_u) * __pyx_v_v)) * ((__pyx_v_verts[3])[__pyx_v_i]))) + (((4.0 * __pyx_v_u) * __pyx_v_v) * ((__pyx_v_verts[4])[__pyx_v_i]))) + ((((4.0 * __pyx_v_v) - ((4.0 * __pyx_v_v) * __pyx_v_v)) - ((4.0 * __pyx_v_u) * __pyx_v_v)) * ((__pyx_v_verts[5])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":405
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceFunc(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                            const cython.floating u,
 *                            const cython.floating v,
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":423
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceDerivU(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU(float (*__pyx_v_verts)[3], float const __pyx_v_u, float const __pyx_v_v, float *__pyx_v_Su) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":429
 *     cdef int i
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Su[i] = (-3.0 + 4.0*u + 4.0*v)*verts[0][i] + \
 *              (-1.0 + 4.0*u)*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":430
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):
 *         Su[i] = (-3.0 + 4.0*u + 4.0*v)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *              (-1.0 + 4.0*u)*verts[1][i] + \
 *              (4.0 - 8.0*u - 4.0*v)*verts[3][i] + \
 */
    (__pyx_v_Su[__pyx_v_i]) = (((((((-3.0 + (4.0 * __pyx_v_u)) + (4.0 * __pyx_v_v)) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((-1.0 + (4.0 * __pyx_v_u)) * ((__pyx_v_verts[1])[__pyx_v_i]))) + (((4.0 - (8.0 * __pyx_v_u)) - (4.0 * __pyx_v_v)) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((4.0 * __pyx_v_v) * ((__pyx_v_verts[4])[__pyx_v_i]))) + ((-4.0 * __pyx_v_v) * ((__pyx_v_verts[5])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":423
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceDerivU(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU(double (*__pyx_v_verts)[3], double const __pyx_v_u, double const __pyx_v_v, double *__pyx_v_Su) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":429
 *     cdef int i
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Su[i] = (-3.0 + 4.0*u + 4.0*v)*verts[0][i] + \
 *              (-1.0 + 4.0*u)*verts[1][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":430
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):
 *         Su[i] = (-3.0 + 4.0*u + 4.0*v)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *              (-1.0 + 4.0*u)*verts[1][i] + \
 *              (4.0 - 8.0*u - 4.0*v)*verts[3][i] + \
 */
    (__pyx_v_Su[__pyx_v_i]) = (((((((-3.0 + (4.0 * __pyx_v_u)) + (4.0 * __pyx_v_v)) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((-1.0 + (4.0 * __pyx_v_u)) * ((__pyx_v_verts[1])[__pyx_v_i]))) + (((4.0 - (8.0 * __pyx_v_u)) - (4.0 * __pyx_v_v)) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((4.0 * __pyx_v_v) * ((__pyx_v_verts[4])[__pyx_v_i]))) + ((-4.0 * __pyx_v_v) * ((__pyx_v_verts[5])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":423
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceDerivU(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":440
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceDerivV(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

static void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV(float (*__pyx_v_verts)[3], float const __pyx_v_u, float const __pyx_v_v, float *__pyx_v_Sv) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":447
 *     cdef int i
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Sv[i] = (-3.0 + 4.0*v + 4.0*u)*verts[0][i] + \
 *              (-1.0 + 4.0*v)*verts[2][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":448
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):
 *         Sv[i] = (-3.0 + 4.0*v + 4.0*u)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *              (-1.0 + 4.0*v)*verts[2][i] + \
 *              (-4.0*u)*verts[3][i] + \
 */
    (__pyx_v_Sv[__pyx_v_i]) = (((((((-3.0 + (4.0 * __pyx_v_v)) + (4.0 * __pyx_v_u)) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((-1.0 + (4.0 * __pyx_v_v)) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((-4.0 * __pyx_v_u) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((4.0 * __pyx_v_u) * ((__pyx_v_verts[4])[__pyx_v_i]))) + (((4.0 - (8.0 * __pyx_v_v)) - (4.0 * __pyx_v_u)) * ((__pyx_v_verts[5])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":440
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceDerivV(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

static void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV(double (*__pyx_v_verts)[3], double const __pyx_v_u, double const __pyx_v_v, double *__pyx_v_Sv) {
  int __pyx_v_i;
  int __pyx_t_1;

  /* "yt/utilities/lib/primitives.pyx":447
 *     cdef int i
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         Sv[i] = (-3.0 + 4.0*v + 4.0*u)*verts[0][i] + \
 *              (-1.0 + 4.0*v)*verts[2][i] + \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":448
 *     # Computes for canonical triangle coordinates
 *     for i in range(3):
 *         Sv[i] = (-3.0 + 4.0*v + 4.0*u)*verts[0][i] + \             # <<<<<<<<<<<<<<
 *              (-1.0 + 4.0*v)*verts[2][i] + \
 *              (-4.0*u)*verts[3][i] + \
 */
    (__pyx_v_Sv[__pyx_v_i]) = (((((((-3.0 + (4.0 * __pyx_v_v)) + (4.0 * __pyx_v_u)) * ((__pyx_v_verts[0])[__pyx_v_i])) + ((-1.0 + (4.0 * __pyx_v_v)) * ((__pyx_v_verts[2])[__pyx_v_i]))) + ((-4.0 * __pyx_v_u) * ((__pyx_v_verts[3])[__pyx_v_i]))) + ((4.0 * __pyx_v_u) * ((__pyx_v_verts[4])[__pyx_v_i]))) + (((4.0 - (8.0 * __pyx_v_v)) - (4.0 * __pyx_v_u)) * ((__pyx_v_verts[5])[__pyx_v_i])));
  }

  /* "yt/utilities/lib/primitives.pyx":440
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patchSurfaceDerivV(const cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                              const cython.floating u,
 *                              const cython.floating v,
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":458
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef RayHitData compute_tet_patch_hit(cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                                   cython.floating[3] ray_origin,
 *                                   cython.floating[3] ray_direction) nogil:
 */

static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit(float (*__pyx_v_verts)[3], float *__pyx_v_ray_origin, float *__pyx_v_ray_direction) {
  float __pyx_v_n[3];
  float __pyx_v_N1[3];
  float __pyx_v_N2[3];
  float __pyx_v_A;
  long __pyx_v_i;
  float __pyx_v_d1;
  float __pyx_v_d2;
  float __pyx_v_u;
  float __pyx_v_v;
  float __pyx_v_S[3];
  float __pyx_v_fu;
  float __pyx_v_fv;
  float __pyx_v_err;
  float __pyx_v_tol;
  int __pyx_v_iterations;
  int __pyx_v_max_iter;
  float __pyx_v_Su[3];
  float __pyx_v_Sv[3];
  float __pyx_v_J11;
  float __pyx_v_J12;
  float __pyx_v_J21;
  float __pyx_v_J22;
  float __pyx_v_det;
  float __pyx_v_t;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_v_hd;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_r;
  long __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":464
 *     # first we compute the two planes that define the ray.
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)             # <<<<<<<<<<<<<<
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A
 */
  __pyx_v_A = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_ray_direction, __pyx_v_ray_direction);

  /* "yt/utilities/lib/primitives.pyx":465
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         n[i] = ray_direction[i] / A
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":466
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A             # <<<<<<<<<<<<<<
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 */
    (__pyx_v_n[__pyx_v_i]) = ((__pyx_v_ray_direction[__pyx_v_i]) / __pyx_v_A);
  }

  /* "yt/utilities/lib/primitives.pyx":468
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[1]))) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[2]))) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/utilities/lib/primitives.pyx":469
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]             # <<<<<<<<<<<<<<
 *         N1[1] =-n[0]
 *         N1[2] = 0.0
 */
    (__pyx_v_N1[0]) = (__pyx_v_n[1]);

    /* "yt/utilities/lib/primitives.pyx":470
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]
 *         N1[1] =-n[0]             # <<<<<<<<<<<<<<
 *         N1[2] = 0.0
 *     else:
 */
    (__pyx_v_N1[1]) = (-(__pyx_v_n[0]));

    /* "yt/utilities/lib/primitives.pyx":471
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 *         N1[2] = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         N1[0] = 0.0
 */
    (__pyx_v_N1[2]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":468
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
    goto __pyx_L5;
  }

  /* "yt/utilities/lib/primitives.pyx":473
 *         N1[2] = 0.0
 *     else:
 *         N1[0] = 0.0             # <<<<<<<<<<<<<<
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 */
  /*else*/ {
    (__pyx_v_N1[0]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":474
 *     else:
 *         N1[0] = 0.0
 *         N1[1] = n[2]             # <<<<<<<<<<<<<<
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)
 */
    (__pyx_v_N1[1]) = (__pyx_v_n[2]);

    /* "yt/utilities/lib/primitives.pyx":475
 *         N1[0] = 0.0
 *         N1[1] = n[2]
 *         N1[2] =-n[1]             # <<<<<<<<<<<<<<
 *     cross(N1, n, N2)
 * 
 */
    (__pyx_v_N1[2]) = (-(__pyx_v_n[1]));
  }
  __pyx_L5:;

  /* "yt/utilities/lib/primitives.pyx":476
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)             # <<<<<<<<<<<<<<
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 */
  __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(__pyx_v_N1, __pyx_v_n, __pyx_v_N2);

  /* "yt/utilities/lib/primitives.pyx":478
 *     cross(N1, n, N2)
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)             # <<<<<<<<<<<<<<
 *     cdef cython.floating d2 = -dot(N2, ray_origin)
 * 
 */
  __pyx_v_d1 = (-__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":479
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 *     cdef cython.floating d2 = -dot(N2, ray_origin)             # <<<<<<<<<<<<<<
 * 
 *     # the initial guess is set to zero
 */
  __pyx_v_d2 = (-__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":482
 * 
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 */
  __pyx_v_u = 0.0;

  /* "yt/utilities/lib/primitives.pyx":483
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0
 *     cdef cython.floating v = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] S
 *     tet_patchSurfaceFunc(verts, u, v, S)
 */
  __pyx_v_v = 0.0;

  /* "yt/utilities/lib/primitives.pyx":485
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 *     tet_patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 */
  __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

  /* "yt/utilities/lib/primitives.pyx":486
 *     cdef cython.floating[3] S
 *     tet_patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 */
  __pyx_v_fu = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

  /* "yt/utilities/lib/primitives.pyx":487
 *     tet_patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 * 
 */
  __pyx_v_fv = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

  /* "yt/utilities/lib/primitives.pyx":488
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 * 
 *     # begin Newton iteration
 */
  __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

  /* "yt/utilities/lib/primitives.pyx":491
 * 
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5             # <<<<<<<<<<<<<<
 *     cdef int iterations = 0
 *     cdef int max_iter = 10
 */
  __pyx_v_tol = 1.0e-5;

  /* "yt/utilities/lib/primitives.pyx":492
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0             # <<<<<<<<<<<<<<
 *     cdef int max_iter = 10
 *     cdef cython.floating[3] Su
 */
  __pyx_v_iterations = 0;

  /* "yt/utilities/lib/primitives.pyx":493
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0
 *     cdef int max_iter = 10             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] Su
 *     cdef cython.floating[3] Sv
 */
  __pyx_v_max_iter = 10;

  /* "yt/utilities/lib/primitives.pyx":497
 *     cdef cython.floating[3] Sv
 *     cdef cython.floating J11, J12, J21, J22, det
 *     while ((err > tol) and (iterations < max_iter)):             # <<<<<<<<<<<<<<
 *         # compute the Jacobian
 *         tet_patchSurfaceDerivU(verts, u, v, Su)
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_err > __pyx_v_tol) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_3 = ((__pyx_v_iterations < __pyx_v_max_iter) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L10_bool_binop_done:;
    if (!__pyx_t_2) break;

    /* "yt/utilities/lib/primitives.pyx":499
 *     while ((err > tol) and (iterations < max_iter)):
 *         # compute the Jacobian
 *         tet_patchSurfaceDerivU(verts, u, v, Su)             # <<<<<<<<<<<<<<
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 */
    __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":500
 *         # compute the Jacobian
 *         tet_patchSurfaceDerivU(verts, u, v, Su)
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)             # <<<<<<<<<<<<<<
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 */
    __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":501
 *         tet_patchSurfaceDerivU(verts, u, v, Su)
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)             # <<<<<<<<<<<<<<
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 */
    __pyx_v_J11 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":502
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)             # <<<<<<<<<<<<<<
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 */
    __pyx_v_J12 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":503
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)             # <<<<<<<<<<<<<<
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)
 */
    __pyx_v_J21 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":504
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)             # <<<<<<<<<<<<<<
 *         det = (J11*J22 - J12*J21)
 * 
 */
    __pyx_v_J22 = __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":505
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)             # <<<<<<<<<<<<<<
 * 
 *         # update the u, v values
 */
    __pyx_v_det = ((__pyx_v_J11 * __pyx_v_J22) - (__pyx_v_J12 * __pyx_v_J21));

    /* "yt/utilities/lib/primitives.pyx":508
 * 
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det             # <<<<<<<<<<<<<<
 *         v -= (-J21*fu + J11*fv) / det
 * 
 */
    __pyx_v_u = (__pyx_v_u - (((__pyx_v_J22 * __pyx_v_fu) - (__pyx_v_J12 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":509
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det
 *         v -= (-J21*fu + J11*fv) / det             # <<<<<<<<<<<<<<
 * 
 *         tet_patchSurfaceFunc(verts, u, v, S)
 */
    __pyx_v_v = (__pyx_v_v - ((((-__pyx_v_J21) * __pyx_v_fu) + (__pyx_v_J11 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":511
 *         v -= (-J21*fu + J11*fv) / det
 * 
 *         tet_patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2
 */
    __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

    /* "yt/utilities/lib/primitives.pyx":512
 * 
 *         tet_patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *         fv = dot(N2, S) + d2
 * 
 */
    __pyx_v_fu = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

    /* "yt/utilities/lib/primitives.pyx":513
 *         tet_patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 */
    __pyx_v_fv = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

    /* "yt/utilities/lib/primitives.pyx":515
 *         fv = dot(N2, S) + d2
 * 
 *         err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 *         iterations += 1
 * 
 */
    __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

    /* "yt/utilities/lib/primitives.pyx":516
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 *         iterations += 1             # <<<<<<<<<<<<<<
 * 
 *     # t is the distance along the ray to this hit
 */
    __pyx_v_iterations = (__pyx_v_iterations + 1);
  }

  /* "yt/utilities/lib/primitives.pyx":519
 * 
 *     # t is the distance along the ray to this hit
 *     cdef cython.floating t = distance(S, ray_origin) / L2_norm(ray_direction)             # <<<<<<<<<<<<<<
 * 
 *     # return hit data
 */
  __pyx_v_t = (__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(__pyx_v_S, __pyx_v_ray_origin) / __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(__pyx_v_ray_direction));

  /* "yt/utilities/lib/primitives.pyx":523
 *     # return hit data
 *     cdef RayHitData hd
 *     hd.u = u             # <<<<<<<<<<<<<<
 *     hd.v = v
 *     hd.t = t
 */
  __pyx_v_hd.u = __pyx_v_u;

  /* "yt/utilities/lib/primitives.pyx":524
 *     cdef RayHitData hd
 *     hd.u = u
 *     hd.v = v             # <<<<<<<<<<<<<<
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 */
  __pyx_v_hd.v = __pyx_v_v;

  /* "yt/utilities/lib/primitives.pyx":525
 *     hd.u = u
 *     hd.v = v
 *     hd.t = t             # <<<<<<<<<<<<<<
 *     hd.converged = (iterations < max_iter)
 *     return hd
 */
  __pyx_v_hd.t = __pyx_v_t;

  /* "yt/utilities/lib/primitives.pyx":526
 *     hd.v = v
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)             # <<<<<<<<<<<<<<
 *     return hd
 * 
 */
  __pyx_v_hd.converged = (__pyx_v_iterations < __pyx_v_max_iter);

  /* "yt/utilities/lib/primitives.pyx":527
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 *     return hd             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_hd;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":458
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef RayHitData compute_tet_patch_hit(cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                                   cython.floating[3] ray_origin,
 *                                   cython.floating[3] ray_direction) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit(double (*__pyx_v_verts)[3], double *__pyx_v_ray_origin, double *__pyx_v_ray_direction) {
  double __pyx_v_n[3];
  double __pyx_v_N1[3];
  double __pyx_v_N2[3];
  double __pyx_v_A;
  long __pyx_v_i;
  double __pyx_v_d1;
  double __pyx_v_d2;
  double __pyx_v_u;
  double __pyx_v_v;
  double __pyx_v_S[3];
  double __pyx_v_fu;
  double __pyx_v_fv;
  double __pyx_v_err;
  double __pyx_v_tol;
  int __pyx_v_iterations;
  int __pyx_v_max_iter;
  double __pyx_v_Su[3];
  double __pyx_v_Sv[3];
  double __pyx_v_J11;
  double __pyx_v_J12;
  double __pyx_v_J21;
  double __pyx_v_J22;
  double __pyx_v_det;
  double __pyx_v_t;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_v_hd;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_r;
  long __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":464
 *     # first we compute the two planes that define the ray.
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)             # <<<<<<<<<<<<<<
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A
 */
  __pyx_v_A = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_ray_direction, __pyx_v_ray_direction);

  /* "yt/utilities/lib/primitives.pyx":465
 *     cdef cython.floating[3] n, N1, N2
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         n[i] = ray_direction[i] / A
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":466
 *     cdef cython.floating A = dot(ray_direction, ray_direction)
 *     for i in range(3):
 *         n[i] = ray_direction[i] / A             # <<<<<<<<<<<<<<
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 */
    (__pyx_v_n[__pyx_v_i]) = ((__pyx_v_ray_direction[__pyx_v_i]) / __pyx_v_A);
  }

  /* "yt/utilities/lib/primitives.pyx":468
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[1]))) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = ((fabs((__pyx_v_n[0])) > fabs((__pyx_v_n[2]))) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/utilities/lib/primitives.pyx":469
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]             # <<<<<<<<<<<<<<
 *         N1[1] =-n[0]
 *         N1[2] = 0.0
 */
    (__pyx_v_N1[0]) = (__pyx_v_n[1]);

    /* "yt/utilities/lib/primitives.pyx":470
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
 *         N1[0] = n[1]
 *         N1[1] =-n[0]             # <<<<<<<<<<<<<<
 *         N1[2] = 0.0
 *     else:
 */
    (__pyx_v_N1[1]) = (-(__pyx_v_n[0]));

    /* "yt/utilities/lib/primitives.pyx":471
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 *         N1[2] = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         N1[0] = 0.0
 */
    (__pyx_v_N1[2]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":468
 *         n[i] = ray_direction[i] / A
 * 
 *     if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):             # <<<<<<<<<<<<<<
 *         N1[0] = n[1]
 *         N1[1] =-n[0]
 */
    goto __pyx_L5;
  }

  /* "yt/utilities/lib/primitives.pyx":473
 *         N1[2] = 0.0
 *     else:
 *         N1[0] = 0.0             # <<<<<<<<<<<<<<
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 */
  /*else*/ {
    (__pyx_v_N1[0]) = 0.0;

    /* "yt/utilities/lib/primitives.pyx":474
 *     else:
 *         N1[0] = 0.0
 *         N1[1] = n[2]             # <<<<<<<<<<<<<<
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)
 */
    (__pyx_v_N1[1]) = (__pyx_v_n[2]);

    /* "yt/utilities/lib/primitives.pyx":475
 *         N1[0] = 0.0
 *         N1[1] = n[2]
 *         N1[2] =-n[1]             # <<<<<<<<<<<<<<
 *     cross(N1, n, N2)
 * 
 */
    (__pyx_v_N1[2]) = (-(__pyx_v_n[1]));
  }
  __pyx_L5:;

  /* "yt/utilities/lib/primitives.pyx":476
 *         N1[1] = n[2]
 *         N1[2] =-n[1]
 *     cross(N1, n, N2)             # <<<<<<<<<<<<<<
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(__pyx_v_N1, __pyx_v_n, __pyx_v_N2);

  /* "yt/utilities/lib/primitives.pyx":478
 *     cross(N1, n, N2)
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)             # <<<<<<<<<<<<<<
 *     cdef cython.floating d2 = -dot(N2, ray_origin)
 * 
 */
  __pyx_v_d1 = (-__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":479
 * 
 *     cdef cython.floating d1 = -dot(N1, ray_origin)
 *     cdef cython.floating d2 = -dot(N2, ray_origin)             # <<<<<<<<<<<<<<
 * 
 *     # the initial guess is set to zero
 */
  __pyx_v_d2 = (-__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_ray_origin));

  /* "yt/utilities/lib/primitives.pyx":482
 * 
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 */
  __pyx_v_u = 0.0;

  /* "yt/utilities/lib/primitives.pyx":483
 *     # the initial guess is set to zero
 *     cdef cython.floating u = 0.0
 *     cdef cython.floating v = 0.0             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] S
 *     tet_patchSurfaceFunc(verts, u, v, S)
 */
  __pyx_v_v = 0.0;

  /* "yt/utilities/lib/primitives.pyx":485
 *     cdef cython.floating v = 0.0
 *     cdef cython.floating[3] S
 *     tet_patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 */
  __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

  /* "yt/utilities/lib/primitives.pyx":486
 *     cdef cython.floating[3] S
 *     tet_patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 */
  __pyx_v_fu = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

  /* "yt/utilities/lib/primitives.pyx":487
 *     tet_patchSurfaceFunc(verts, u, v, S)
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))
 * 
 */
  __pyx_v_fv = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

  /* "yt/utilities/lib/primitives.pyx":488
 *     cdef cython.floating fu = dot(N1, S) + d1
 *     cdef cython.floating fv = dot(N2, S) + d2
 *     cdef cython.floating err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 * 
 *     # begin Newton iteration
 */
  __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

  /* "yt/utilities/lib/primitives.pyx":491
 * 
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5             # <<<<<<<<<<<<<<
 *     cdef int iterations = 0
 *     cdef int max_iter = 10
 */
  __pyx_v_tol = 1.0e-5;

  /* "yt/utilities/lib/primitives.pyx":492
 *     # begin Newton iteration
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0             # <<<<<<<<<<<<<<
 *     cdef int max_iter = 10
 *     cdef cython.floating[3] Su
 */
  __pyx_v_iterations = 0;

  /* "yt/utilities/lib/primitives.pyx":493
 *     cdef cython.floating tol = 1.0e-5
 *     cdef int iterations = 0
 *     cdef int max_iter = 10             # <<<<<<<<<<<<<<
 *     cdef cython.floating[3] Su
 *     cdef cython.floating[3] Sv
 */
  __pyx_v_max_iter = 10;

  /* "yt/utilities/lib/primitives.pyx":497
 *     cdef cython.floating[3] Sv
 *     cdef cython.floating J11, J12, J21, J22, det
 *     while ((err > tol) and (iterations < max_iter)):             # <<<<<<<<<<<<<<
 *         # compute the Jacobian
 *         tet_patchSurfaceDerivU(verts, u, v, Su)
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_err > __pyx_v_tol) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_3 = ((__pyx_v_iterations < __pyx_v_max_iter) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L10_bool_binop_done:;
    if (!__pyx_t_2) break;

    /* "yt/utilities/lib/primitives.pyx":499
 *     while ((err > tol) and (iterations < max_iter)):
 *         # compute the Jacobian
 *         tet_patchSurfaceDerivU(verts, u, v, Su)             # <<<<<<<<<<<<<<
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 */
    __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":500
 *         # compute the Jacobian
 *         tet_patchSurfaceDerivU(verts, u, v, Su)
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)             # <<<<<<<<<<<<<<
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 */
    __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":501
 *         tet_patchSurfaceDerivU(verts, u, v, Su)
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)             # <<<<<<<<<<<<<<
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 */
    __pyx_v_J11 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":502
 *         tet_patchSurfaceDerivV(verts, u, v, Sv)
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)             # <<<<<<<<<<<<<<
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 */
    __pyx_v_J12 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":503
 *         J11 = dot(N1, Su)
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)             # <<<<<<<<<<<<<<
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)
 */
    __pyx_v_J21 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Su);

    /* "yt/utilities/lib/primitives.pyx":504
 *         J12 = dot(N1, Sv)
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)             # <<<<<<<<<<<<<<
 *         det = (J11*J22 - J12*J21)
 * 
 */
    __pyx_v_J22 = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_Sv);

    /* "yt/utilities/lib/primitives.pyx":505
 *         J21 = dot(N2, Su)
 *         J22 = dot(N2, Sv)
 *         det = (J11*J22 - J12*J21)             # <<<<<<<<<<<<<<
 * 
 *         # update the u, v values
 */
    __pyx_v_det = ((__pyx_v_J11 * __pyx_v_J22) - (__pyx_v_J12 * __pyx_v_J21));

    /* "yt/utilities/lib/primitives.pyx":508
 * 
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det             # <<<<<<<<<<<<<<
 *         v -= (-J21*fu + J11*fv) / det
 * 
 */
    __pyx_v_u = (__pyx_v_u - (((__pyx_v_J22 * __pyx_v_fu) - (__pyx_v_J12 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":509
 *         # update the u, v values
 *         u -= ( J22*fu - J12*fv) / det
 *         v -= (-J21*fu + J11*fv) / det             # <<<<<<<<<<<<<<
 * 
 *         tet_patchSurfaceFunc(verts, u, v, S)
 */
    __pyx_v_v = (__pyx_v_v - ((((-__pyx_v_J21) * __pyx_v_fu) + (__pyx_v_J11 * __pyx_v_fv)) / __pyx_v_det));

    /* "yt/utilities/lib/primitives.pyx":511
 *         v -= (-J21*fu + J11*fv) / det
 * 
 *         tet_patchSurfaceFunc(verts, u, v, S)             # <<<<<<<<<<<<<<
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2
 */
    __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc(__pyx_v_verts, __pyx_v_u, __pyx_v_v, __pyx_v_S);

    /* "yt/utilities/lib/primitives.pyx":512
 * 
 *         tet_patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1             # <<<<<<<<<<<<<<
 *         fv = dot(N2, S) + d2
 * 
 */
    __pyx_v_fu = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N1, __pyx_v_S) + __pyx_v_d1);

    /* "yt/utilities/lib/primitives.pyx":513
 *         tet_patchSurfaceFunc(verts, u, v, S)
 *         fu = dot(N1, S) + d1
 *         fv = dot(N2, S) + d2             # <<<<<<<<<<<<<<
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 */
    __pyx_v_fv = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(__pyx_v_N2, __pyx_v_S) + __pyx_v_d2);

    /* "yt/utilities/lib/primitives.pyx":515
 *         fv = dot(N2, S) + d2
 * 
 *         err = fmax(fabs(fu), fabs(fv))             # <<<<<<<<<<<<<<
 *         iterations += 1
 * 
 */
    __pyx_v_err = fmax(fabs(__pyx_v_fu), fabs(__pyx_v_fv));

    /* "yt/utilities/lib/primitives.pyx":516
 * 
 *         err = fmax(fabs(fu), fabs(fv))
 *         iterations += 1             # <<<<<<<<<<<<<<
 * 
 *     # t is the distance along the ray to this hit
 */
    __pyx_v_iterations = (__pyx_v_iterations + 1);
  }

  /* "yt/utilities/lib/primitives.pyx":519
 * 
 *     # t is the distance along the ray to this hit
 *     cdef cython.floating t = distance(S, ray_origin) / L2_norm(ray_direction)             # <<<<<<<<<<<<<<
 * 
 *     # return hit data
 */
  __pyx_v_t = (__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(__pyx_v_S, __pyx_v_ray_origin) / __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(__pyx_v_ray_direction));

  /* "yt/utilities/lib/primitives.pyx":523
 *     # return hit data
 *     cdef RayHitData hd
 *     hd.u = u             # <<<<<<<<<<<<<<
 *     hd.v = v
 *     hd.t = t
 */
  __pyx_v_hd.u = __pyx_v_u;

  /* "yt/utilities/lib/primitives.pyx":524
 *     cdef RayHitData hd
 *     hd.u = u
 *     hd.v = v             # <<<<<<<<<<<<<<
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 */
  __pyx_v_hd.v = __pyx_v_v;

  /* "yt/utilities/lib/primitives.pyx":525
 *     hd.u = u
 *     hd.v = v
 *     hd.t = t             # <<<<<<<<<<<<<<
 *     hd.converged = (iterations < max_iter)
 *     return hd
 */
  __pyx_v_hd.t = __pyx_v_t;

  /* "yt/utilities/lib/primitives.pyx":526
 *     hd.v = v
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)             # <<<<<<<<<<<<<<
 *     return hd
 * 
 */
  __pyx_v_hd.converged = (__pyx_v_iterations < __pyx_v_max_iter);

  /* "yt/utilities/lib/primitives.pyx":527
 *     hd.t = t
 *     hd.converged = (iterations < max_iter)
 *     return hd             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_hd;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":458
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef RayHitData compute_tet_patch_hit(cython.floating[6][3] verts,             # <<<<<<<<<<<<<<
 *                                   cython.floating[3] ray_origin,
 *                                   cython.floating[3] ray_direction) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/primitives.pyx":533
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_tet_patch_intersect(const void* primitives,             # <<<<<<<<<<<<<<
 *                                     const np.int64_t item,
 *                                     Ray* ray) nogil:
 */

static __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_10primitives_ray_tet_patch_intersect(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *__pyx_v_ray) {
  struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch __pyx_v_tet_patch;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData __pyx_v_hd;
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5numpy_float64_t __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;

  /* "yt/utilities/lib/primitives.pyx":537
 *                                     Ray* ray) nogil:
 * 
 *     cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     cdef RayHitData hd = compute_tet_patch_hit(tet_patch.v, ray.origin, ray.direction)
 */
  __pyx_v_tet_patch = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":539
 *     cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]
 * 
 *     cdef RayHitData hd = compute_tet_patch_hit(tet_patch.v, ray.origin, ray.direction)             # <<<<<<<<<<<<<<
 * 
 *     # only count this is it's the closest hit
 */
  __pyx_v_hd = __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit(__pyx_v_tet_patch.v, __pyx_v_ray->origin, __pyx_v_ray->direction);

  /* "yt/utilities/lib/primitives.pyx":542
 * 
 *     # only count this is it's the closest hit
 *     if (hd.t < ray.t_near or hd.t > ray.t_far):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = ((__pyx_v_hd.t < __pyx_v_ray->t_near) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_hd.t > __pyx_v_ray->t_far) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":543
 *     # only count this is it's the closest hit
 *     if (hd.t < ray.t_near or hd.t > ray.t_far):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     if (0 <= hd.u and 0 <= hd.v and hd.u + hd.v <= 1.0 and hd.converged):
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":542
 * 
 *     # only count this is it's the closest hit
 *     if (hd.t < ray.t_near or hd.t > ray.t_far):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "yt/utilities/lib/primitives.pyx":545
 *         return False
 * 
 *     if (0 <= hd.u and 0 <= hd.v and hd.u + hd.v <= 1.0 and hd.converged):             # <<<<<<<<<<<<<<
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t
 */
  __pyx_t_2 = ((0.0 <= __pyx_v_hd.u) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((0.0 <= __pyx_v_hd.v) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_hd.u + __pyx_v_hd.v) <= 1.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_hd.converged != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/utilities/lib/primitives.pyx":547
 *     if (0 <= hd.u and 0 <= hd.v and hd.u + hd.v <= 1.0 and hd.converged):
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t             # <<<<<<<<<<<<<<
 *         ray.elem_id = tet_patch.elem_id
 *         return True
 */
    __pyx_t_3 = __pyx_v_hd.t;
    __pyx_v_ray->t_far = __pyx_t_3;

    /* "yt/utilities/lib/primitives.pyx":548
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t
 *         ray.elem_id = tet_patch.elem_id             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_t_4 = __pyx_v_tet_patch.elem_id;
    __pyx_v_ray->elem_id = __pyx_t_4;

    /* "yt/utilities/lib/primitives.pyx":549
 *         ray.t_far = hd.t
 *         ray.elem_id = tet_patch.elem_id
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     return False
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/utilities/lib/primitives.pyx":545
 *         return False
 * 
 *     if (0 <= hd.u and 0 <= hd.v and hd.u + hd.v <= 1.0 and hd.converged):             # <<<<<<<<<<<<<<
 *         # we have a hit, so update ray information
 *         ray.t_far = hd.t
 */
  }

  /* "yt/utilities/lib/primitives.pyx":551
 *         return True
 * 
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/utilities/lib/primitives.pyx":533
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef np.int64_t ray_tet_patch_intersect(const void* primitives,             # <<<<<<<<<<<<<<
 *                                     const np.int64_t item,
 *                                     Ray* ray) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/primitives.pyx":557
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patch_centroid(const void *primitives,             # <<<<<<<<<<<<<<
 *                          const np.int64_t item,
 *                          np.float64_t[3] centroid) nogil:
 */

static void __pyx_f_2yt_9utilities_3lib_10primitives_tet_patch_centroid(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, __pyx_t_5numpy_float64_t *__pyx_v_centroid) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_j;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch __pyx_v_tet_patch;
  __pyx_t_5numpy_int64_t __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":562
 * 
 *     cdef np.int64_t i, j
 *     cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(3):
 */
  __pyx_v_tet_patch = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":564
 *     cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]
 * 
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         centroid[j] = 0.0
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":565
 * 
 *     for j in range(3):
 *         centroid[j] = 0.0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(6):
 */
    (__pyx_v_centroid[__pyx_v_j]) = 0.0;
  }

  /* "yt/utilities/lib/primitives.pyx":567
 *         centroid[j] = 0.0
 * 
 *     for i in range(6):             # <<<<<<<<<<<<<<
 *         for j in range(3):
 *             centroid[j] += tet_patch.v[i][j]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 6; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":568
 * 
 *     for i in range(6):
 *         for j in range(3):             # <<<<<<<<<<<<<<
 *             centroid[j] += tet_patch.v[i][j]
 * 
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/utilities/lib/primitives.pyx":569
 *     for i in range(6):
 *         for j in range(3):
 *             centroid[j] += tet_patch.v[i][j]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(3):
 */
      __pyx_t_3 = __pyx_v_j;
      (__pyx_v_centroid[__pyx_t_3]) = ((__pyx_v_centroid[__pyx_t_3]) + ((__pyx_v_tet_patch.v[__pyx_v_i])[__pyx_v_j]));
    }
  }

  /* "yt/utilities/lib/primitives.pyx":571
 *             centroid[j] += tet_patch.v[i][j]
 * 
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         centroid[j] /= 6.0
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":572
 * 
 *     for j in range(3):
 *         centroid[j] /= 6.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_v_j;
    (__pyx_v_centroid[__pyx_t_2]) = ((__pyx_v_centroid[__pyx_t_2]) / 6.0);
  }

  /* "yt/utilities/lib/primitives.pyx":557
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patch_centroid(const void *primitives,             # <<<<<<<<<<<<<<
 *                          const np.int64_t item,
 *                          np.float64_t[3] centroid) nogil:
 */

  /* function exit code */
}

/* "yt/utilities/lib/primitives.pyx":578
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patch_bbox(const void *primitives,             # <<<<<<<<<<<<<<
 *                     const np.int64_t item,
 *                      BBox* bbox) nogil:
 */

static void __pyx_f_2yt_9utilities_3lib_10primitives_tet_patch_bbox(void const *__pyx_v_primitives, __pyx_t_5numpy_int64_t const __pyx_v_item, struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox *__pyx_v_bbox) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_j;
  struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch __pyx_v_tet_patch;
  __pyx_t_5numpy_int64_t __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;

  /* "yt/utilities/lib/primitives.pyx":583
 * 
 *     cdef np.int64_t i, j
 *     cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(3):
 */
  __pyx_v_tet_patch = (((struct __pyx_t_2yt_9utilities_3lib_10primitives_TetPatch *)__pyx_v_primitives)[__pyx_v_item]);

  /* "yt/utilities/lib/primitives.pyx":585
 *     cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]
 * 
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         bbox.left_edge[j] = tet_patch.v[0][j]
 *         bbox.right_edge[j] = tet_patch.v[0][j]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":586
 * 
 *     for j in range(3):
 *         bbox.left_edge[j] = tet_patch.v[0][j]             # <<<<<<<<<<<<<<
 *         bbox.right_edge[j] = tet_patch.v[0][j]
 * 
 */
    (__pyx_v_bbox->left_edge[__pyx_v_j]) = ((__pyx_v_tet_patch.v[0])[__pyx_v_j]);

    /* "yt/utilities/lib/primitives.pyx":587
 *     for j in range(3):
 *         bbox.left_edge[j] = tet_patch.v[0][j]
 *         bbox.right_edge[j] = tet_patch.v[0][j]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(1, 6):
 */
    (__pyx_v_bbox->right_edge[__pyx_v_j]) = ((__pyx_v_tet_patch.v[0])[__pyx_v_j]);
  }

  /* "yt/utilities/lib/primitives.pyx":589
 *         bbox.right_edge[j] = tet_patch.v[0][j]
 * 
 *     for i in range(1, 6):             # <<<<<<<<<<<<<<
 *         for j in range(3):
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], tet_patch.v[i][j])
 */
  for (__pyx_t_1 = 1; __pyx_t_1 < 6; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/primitives.pyx":590
 * 
 *     for i in range(1, 6):
 *         for j in range(3):             # <<<<<<<<<<<<<<
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], tet_patch.v[i][j])
 *             bbox.right_edge[j] = fmax(bbox.right_edge[j], tet_patch.v[i][j])
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/utilities/lib/primitives.pyx":591
 *     for i in range(1, 6):
 *         for j in range(3):
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], tet_patch.v[i][j])             # <<<<<<<<<<<<<<
 *             bbox.right_edge[j] = fmax(bbox.right_edge[j], tet_patch.v[i][j])
 */
      (__pyx_v_bbox->left_edge[__pyx_v_j]) = fmin((__pyx_v_bbox->left_edge[__pyx_v_j]), ((__pyx_v_tet_patch.v[__pyx_v_i])[__pyx_v_j]));

      /* "yt/utilities/lib/primitives.pyx":592
 *         for j in range(3):
 *             bbox.left_edge[j] = fmin(bbox.left_edge[j], tet_patch.v[i][j])
 *             bbox.right_edge[j] = fmax(bbox.right_edge[j], tet_patch.v[i][j])             # <<<<<<<<<<<<<<
 */
      (__pyx_v_bbox->right_edge[__pyx_v_j]) = fmax((__pyx_v_bbox->right_edge[__pyx_v_j]), ((__pyx_v_tet_patch.v[__pyx_v_i])[__pyx_v_j]));
    }
  }

  /* "yt/utilities/lib/primitives.pyx":578
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef void tet_patch_bbox(const void *primitives,             # <<<<<<<<<<<<<<
 *                     const np.int64_t item,
 *                      BBox* bbox) nogil:
 */

  /* function exit code */
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":931
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":937
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":943
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L3_error)

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 945, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 945, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":949
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 949, __pyx_L3_error)

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 951, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 951, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":955
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 955, __pyx_L3_error)

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":957
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 957, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 957, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":979
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":994
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1004
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1011
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1018
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "vec3_ops.pxd":9
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating dot(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                                 const cython.floating[3] b) nogil:
 *     return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]
 */

static CYTHON_INLINE float __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(float *__pyx_v_a, float *__pyx_v_b) {
  float __pyx_r;

  /* "vec3_ops.pxd":11
 * cdef inline cython.floating dot(const cython.floating[3] a,
 *                                 const cython.floating[3] b) nogil:
 *     return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((((__pyx_v_a[0]) * (__pyx_v_b[0])) + ((__pyx_v_a[1]) * (__pyx_v_b[1]))) + ((__pyx_v_a[2]) * (__pyx_v_b[2])));
  goto __pyx_L0;

  /* "vec3_ops.pxd":9
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating dot(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                                 const cython.floating[3] b) nogil:
 *     return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE double __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_dot(double *__pyx_v_a, double *__pyx_v_b) {
  double __pyx_r;

  /* "vec3_ops.pxd":11
 * cdef inline cython.floating dot(const cython.floating[3] a,
 *                                 const cython.floating[3] b) nogil:
 *     return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((((__pyx_v_a[0]) * (__pyx_v_b[0])) + ((__pyx_v_a[1]) * (__pyx_v_b[1]))) + ((__pyx_v_a[2]) * (__pyx_v_b[2])));
  goto __pyx_L0;

  /* "vec3_ops.pxd":9
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating dot(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                                 const cython.floating[3] b) nogil:
 *     return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "vec3_ops.pxd":17
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void cross(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                        const cython.floating[3] b,
 *                        cython.floating c[3]) nogil:
 */

static CYTHON_INLINE void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(float *__pyx_v_a, float *__pyx_v_b, float *__pyx_v_c) {

  /* "vec3_ops.pxd":20
 *                        const cython.floating[3] b,
 *                        cython.floating c[3]) nogil:
 *     c[0] = a[1]*b[2] - a[2]*b[1]             # <<<<<<<<<<<<<<
 *     c[1] = a[2]*b[0] - a[0]*b[2]
 *     c[2] = a[0]*b[1] - a[1]*b[0]
 */
  (__pyx_v_c[0]) = (((__pyx_v_a[1]) * (__pyx_v_b[2])) - ((__pyx_v_a[2]) * (__pyx_v_b[1])));

  /* "vec3_ops.pxd":21
 *                        cython.floating c[3]) nogil:
 *     c[0] = a[1]*b[2] - a[2]*b[1]
 *     c[1] = a[2]*b[0] - a[0]*b[2]             # <<<<<<<<<<<<<<
 *     c[2] = a[0]*b[1] - a[1]*b[0]
 * 
 */
  (__pyx_v_c[1]) = (((__pyx_v_a[2]) * (__pyx_v_b[0])) - ((__pyx_v_a[0]) * (__pyx_v_b[2])));

  /* "vec3_ops.pxd":22
 *     c[0] = a[1]*b[2] - a[2]*b[1]
 *     c[1] = a[2]*b[0] - a[0]*b[2]
 *     c[2] = a[0]*b[1] - a[1]*b[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_c[2]) = (((__pyx_v_a[0]) * (__pyx_v_b[1])) - ((__pyx_v_a[1]) * (__pyx_v_b[0])));

  /* "vec3_ops.pxd":17
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void cross(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                        const cython.floating[3] b,
 *                        cython.floating c[3]) nogil:
 */

  /* function exit code */
}

static CYTHON_INLINE void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_cross(double *__pyx_v_a, double *__pyx_v_b, double *__pyx_v_c) {

  /* "vec3_ops.pxd":20
 *                        const cython.floating[3] b,
 *                        cython.floating c[3]) nogil:
 *     c[0] = a[1]*b[2] - a[2]*b[1]             # <<<<<<<<<<<<<<
 *     c[1] = a[2]*b[0] - a[0]*b[2]
 *     c[2] = a[0]*b[1] - a[1]*b[0]
 */
  (__pyx_v_c[0]) = (((__pyx_v_a[1]) * (__pyx_v_b[2])) - ((__pyx_v_a[2]) * (__pyx_v_b[1])));

  /* "vec3_ops.pxd":21
 *                        cython.floating c[3]) nogil:
 *     c[0] = a[1]*b[2] - a[2]*b[1]
 *     c[1] = a[2]*b[0] - a[0]*b[2]             # <<<<<<<<<<<<<<
 *     c[2] = a[0]*b[1] - a[1]*b[0]
 * 
 */
  (__pyx_v_c[1]) = (((__pyx_v_a[2]) * (__pyx_v_b[0])) - ((__pyx_v_a[0]) * (__pyx_v_b[2])));

  /* "vec3_ops.pxd":22
 *     c[0] = a[1]*b[2] - a[2]*b[1]
 *     c[1] = a[2]*b[0] - a[0]*b[2]
 *     c[2] = a[0]*b[1] - a[1]*b[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_c[2]) = (((__pyx_v_a[0]) * (__pyx_v_b[1])) - ((__pyx_v_a[1]) * (__pyx_v_b[0])));

  /* "vec3_ops.pxd":17
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void cross(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                        const cython.floating[3] b,
 *                        cython.floating c[3]) nogil:
 */

  /* function exit code */
}

/* "vec3_ops.pxd":28
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void subtract(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                           const cython.floating[3] b,
 *                           cython.floating c[3]) nogil:
 */

static CYTHON_INLINE void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_subtract(float *__pyx_v_a, float *__pyx_v_b, float *__pyx_v_c) {

  /* "vec3_ops.pxd":31
 *                           const cython.floating[3] b,
 *                           cython.floating c[3]) nogil:
 *     c[0] = a[0] - b[0]             # <<<<<<<<<<<<<<
 *     c[1] = a[1] - b[1]
 *     c[2] = a[2] - b[2]
 */
  (__pyx_v_c[0]) = ((__pyx_v_a[0]) - (__pyx_v_b[0]));

  /* "vec3_ops.pxd":32
 *                           cython.floating c[3]) nogil:
 *     c[0] = a[0] - b[0]
 *     c[1] = a[1] - b[1]             # <<<<<<<<<<<<<<
 *     c[2] = a[2] - b[2]
 * 
 */
  (__pyx_v_c[1]) = ((__pyx_v_a[1]) - (__pyx_v_b[1]));

  /* "vec3_ops.pxd":33
 *     c[0] = a[0] - b[0]
 *     c[1] = a[1] - b[1]
 *     c[2] = a[2] - b[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_c[2]) = ((__pyx_v_a[2]) - (__pyx_v_b[2]));

  /* "vec3_ops.pxd":28
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void subtract(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                           const cython.floating[3] b,
 *                           cython.floating c[3]) nogil:
 */

  /* function exit code */
}

static CYTHON_INLINE void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_subtract(double *__pyx_v_a, double *__pyx_v_b, double *__pyx_v_c) {

  /* "vec3_ops.pxd":31
 *                           const cython.floating[3] b,
 *                           cython.floating c[3]) nogil:
 *     c[0] = a[0] - b[0]             # <<<<<<<<<<<<<<
 *     c[1] = a[1] - b[1]
 *     c[2] = a[2] - b[2]
 */
  (__pyx_v_c[0]) = ((__pyx_v_a[0]) - (__pyx_v_b[0]));

  /* "vec3_ops.pxd":32
 *                           cython.floating c[3]) nogil:
 *     c[0] = a[0] - b[0]
 *     c[1] = a[1] - b[1]             # <<<<<<<<<<<<<<
 *     c[2] = a[2] - b[2]
 * 
 */
  (__pyx_v_c[1]) = ((__pyx_v_a[1]) - (__pyx_v_b[1]));

  /* "vec3_ops.pxd":33
 *     c[0] = a[0] - b[0]
 *     c[1] = a[1] - b[1]
 *     c[2] = a[2] - b[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_c[2]) = ((__pyx_v_a[2]) - (__pyx_v_b[2]));

  /* "vec3_ops.pxd":28
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void subtract(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                           const cython.floating[3] b,
 *                           cython.floating c[3]) nogil:
 */

  /* function exit code */
}

/* "vec3_ops.pxd":39
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating distance(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                                      const cython.floating[3] b) nogil:
 *     return sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 +(a[2] - b[2])**2)
 */

static CYTHON_INLINE float __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(float *__pyx_v_a, float *__pyx_v_b) {
  float __pyx_r;

  /* "vec3_ops.pxd":41
 * cdef inline cython.floating distance(const cython.floating[3] a,
 *                                      const cython.floating[3] b) nogil:
 *     return sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 +(a[2] - b[2])**2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = sqrt(((powf(((__pyx_v_a[0]) - (__pyx_v_b[0])), 2.0) + powf(((__pyx_v_a[1]) - (__pyx_v_b[1])), 2.0)) + powf(((__pyx_v_a[2]) - (__pyx_v_b[2])), 2.0)));
  goto __pyx_L0;

  /* "vec3_ops.pxd":39
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating distance(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                                      const cython.floating[3] b) nogil:
 *     return sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 +(a[2] - b[2])**2)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE double __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_distance(double *__pyx_v_a, double *__pyx_v_b) {
  double __pyx_r;

  /* "vec3_ops.pxd":41
 * cdef inline cython.floating distance(const cython.floating[3] a,
 *                                      const cython.floating[3] b) nogil:
 *     return sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 +(a[2] - b[2])**2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = sqrt(((pow(((__pyx_v_a[0]) - (__pyx_v_b[0])), 2.0) + pow(((__pyx_v_a[1]) - (__pyx_v_b[1])), 2.0)) + pow(((__pyx_v_a[2]) - (__pyx_v_b[2])), 2.0)));
  goto __pyx_L0;

  /* "vec3_ops.pxd":39
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating distance(const cython.floating[3] a,             # <<<<<<<<<<<<<<
 *                                      const cython.floating[3] b) nogil:
 *     return sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 +(a[2] - b[2])**2)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "vec3_ops.pxd":47
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void fma(const cython.floating f,             # <<<<<<<<<<<<<<
 *                      const cython.floating[3] a,
 *                      const cython.floating[3] b,
 */

static CYTHON_INLINE void __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_fma(float const __pyx_v_f, float *__pyx_v_a, float *__pyx_v_b, float *__pyx_v_c) {

  /* "vec3_ops.pxd":51
 *                      const cython.floating[3] b,
 *                      cython.floating[3] c) nogil:
 *     c[0] = f * a[0] + b[0]             # <<<<<<<<<<<<<<
 *     c[1] = f * a[1] + b[1]
 *     c[2] = f * a[2] + b[2]
 */
  (__pyx_v_c[0]) = ((__pyx_v_f * (__pyx_v_a[0])) + (__pyx_v_b[0]));

  /* "vec3_ops.pxd":52
 *                      cython.floating[3] c) nogil:
 *     c[0] = f * a[0] + b[0]
 *     c[1] = f * a[1] + b[1]             # <<<<<<<<<<<<<<
 *     c[2] = f * a[2] + b[2]
 * 
 */
  (__pyx_v_c[1]) = ((__pyx_v_f * (__pyx_v_a[1])) + (__pyx_v_b[1]));

  /* "vec3_ops.pxd":53
 *     c[0] = f * a[0] + b[0]
 *     c[1] = f * a[1] + b[1]
 *     c[2] = f * a[2] + b[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_c[2]) = ((__pyx_v_f * (__pyx_v_a[2])) + (__pyx_v_b[2]));

  /* "vec3_ops.pxd":47
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void fma(const cython.floating f,             # <<<<<<<<<<<<<<
 *                      const cython.floating[3] a,
 *                      const cython.floating[3] b,
 */

  /* function exit code */
}

static CYTHON_INLINE void __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_fma(double const __pyx_v_f, double *__pyx_v_a, double *__pyx_v_b, double *__pyx_v_c) {

  /* "vec3_ops.pxd":51
 *                      const cython.floating[3] b,
 *                      cython.floating[3] c) nogil:
 *     c[0] = f * a[0] + b[0]             # <<<<<<<<<<<<<<
 *     c[1] = f * a[1] + b[1]
 *     c[2] = f * a[2] + b[2]
 */
  (__pyx_v_c[0]) = ((__pyx_v_f * (__pyx_v_a[0])) + (__pyx_v_b[0]));

  /* "vec3_ops.pxd":52
 *                      cython.floating[3] c) nogil:
 *     c[0] = f * a[0] + b[0]
 *     c[1] = f * a[1] + b[1]             # <<<<<<<<<<<<<<
 *     c[2] = f * a[2] + b[2]
 * 
 */
  (__pyx_v_c[1]) = ((__pyx_v_f * (__pyx_v_a[1])) + (__pyx_v_b[1]));

  /* "vec3_ops.pxd":53
 *     c[0] = f * a[0] + b[0]
 *     c[1] = f * a[1] + b[1]
 *     c[2] = f * a[2] + b[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_c[2]) = ((__pyx_v_f * (__pyx_v_a[2])) + (__pyx_v_b[2]));

  /* "vec3_ops.pxd":47
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline void fma(const cython.floating f,             # <<<<<<<<<<<<<<
 *                      const cython.floating[3] a,
 *                      const cython.floating[3] b,
 */

  /* function exit code */
}

/* "vec3_ops.pxd":59
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating L2_norm(const cython.floating[3] a) nogil:             # <<<<<<<<<<<<<<
 *     return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])
 */

static CYTHON_INLINE float __pyx_fuse_0__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(float *__pyx_v_a) {
  float __pyx_r;

  /* "vec3_ops.pxd":60
 * @cython.cdivision(True)
 * cdef inline cython.floating L2_norm(const cython.floating[3] a) nogil:
 *     return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])             # <<<<<<<<<<<<<<
 */
  __pyx_r = sqrt(((((__pyx_v_a[0]) * (__pyx_v_a[0])) + ((__pyx_v_a[1]) * (__pyx_v_a[1]))) + ((__pyx_v_a[2]) * (__pyx_v_a[2]))));
  goto __pyx_L0;

  /* "vec3_ops.pxd":59
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating L2_norm(const cython.floating[3] a) nogil:             # <<<<<<<<<<<<<<
 *     return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE double __pyx_fuse_1__pyx_f_2yt_9utilities_3lib_8vec3_ops_L2_norm(double *__pyx_v_a) {
  double __pyx_r;

  /* "vec3_ops.pxd":60
 * @cython.cdivision(True)
 * cdef inline cython.floating L2_norm(const cython.floating[3] a) nogil:
 *     return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])             # <<<<<<<<<<<<<<
 */
  __pyx_r = sqrt(((((__pyx_v_a[0]) * (__pyx_v_a[0])) + ((__pyx_v_a[1]) * (__pyx_v_a[1]))) + ((__pyx_v_a[2]) * (__pyx_v_a[2]))));
  goto __pyx_L0;

  /* "vec3_ops.pxd":59
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating L2_norm(const cython.floating[3] a) nogil:             # <<<<<<<<<<<<<<
 *     return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_primitives(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_primitives},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "primitives",
    __pyx_k_This_file_contains_definitions, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 945, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../conda-py3/envs/py39/lib/python3.9/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("ray_bbox_intersect", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_ray_bbox_intersect, "__pyx_t_5numpy_int64_t (struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *, struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox const )") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ray_triangle_intersect", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_ray_triangle_intersect, "__pyx_t_5numpy_int64_t (void const *, __pyx_t_5numpy_int64_t const , struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("triangle_centroid", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_triangle_centroid, "void (void const *, __pyx_t_5numpy_int64_t const , __pyx_t_5numpy_float64_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("triangle_bbox", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_triangle_bbox, "void (void const *, __pyx_t_5numpy_int64_t const , struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ray_patch_intersect", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_ray_patch_intersect, "__pyx_t_5numpy_int64_t (void const *, __pyx_t_5numpy_int64_t const , struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("patch_centroid", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_patch_centroid, "void (void const *, __pyx_t_5numpy_int64_t const , __pyx_t_5numpy_float64_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("patch_bbox", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_patch_bbox, "void (void const *, __pyx_t_5numpy_int64_t const , struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ray_tet_patch_intersect", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_ray_tet_patch_intersect, "__pyx_t_5numpy_int64_t (void const *, __pyx_t_5numpy_int64_t const , struct __pyx_t_2yt_9utilities_3lib_10primitives_Ray *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("tet_patch_centroid", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_tet_patch_centroid, "void (void const *, __pyx_t_5numpy_int64_t const , __pyx_t_5numpy_float64_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("tet_patch_bbox", (void (*)(void))__pyx_f_2yt_9utilities_3lib_10primitives_tet_patch_bbox, "void (void const *, __pyx_t_5numpy_int64_t const , struct __pyx_t_2yt_9utilities_3lib_10primitives_BBox *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0patchSurfaceFunc", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc, "void (float (*)[3], float const , float const , float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1patchSurfaceFunc", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceFunc, "void (double (*)[3], double const , double const , double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0patchSurfaceDerivU", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU, "void (float (*)[3], float const , float const , float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1patchSurfaceDerivU", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivU, "void (double (*)[3], double const , double const , double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0patchSurfaceDerivV", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV, "void (float (*)[3], float const , float const , float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1patchSurfaceDerivV", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_patchSurfaceDerivV, "void (double (*)[3], double const , double const , double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0compute_patch_hit", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit, "struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData (float (*)[3], float *, float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1compute_patch_hit", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_patch_hit, "struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData (double (*)[3], double *, double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0compute_tet_patch_hit", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit, "struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData (float (*)[3], float *, float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1compute_tet_patch_hit", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_compute_tet_patch_hit, "struct __pyx_t_2yt_9utilities_3lib_10primitives_RayHitData (double (*)[3], double *, double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0tet_patchSurfaceFunc", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc, "void (float (*)[3], float const , float const , float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1tet_patchSurfaceFunc", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceFunc, "void (double (*)[3], double const , double const , double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0tet_patchSurfaceDerivU", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU, "void (float (*)[3], float const , float const , float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1tet_patchSurfaceDerivU", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivU, "void (double (*)[3], double const , double const , double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0tet_patchSurfaceDerivV", (void (*)(void))__pyx_fuse_0__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV, "void (float (*)[3], float const , float const , float *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1tet_patchSurfaceDerivV", (void (*)(void))__pyx_fuse_1__pyx_f_2yt_9utilities_3lib_10primitives_tet_patchSurfaceDerivV, "void (double (*)[3], double const , double const , double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initprimitives(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initprimitives(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_primitives(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_primitives(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_primitives(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'primitives' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_primitives(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("primitives", __pyx_methods, __pyx_k_This_file_contains_definitions, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_yt__utilities__lib__primitives) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "yt.utilities.lib.primitives")) {
      if (unlikely(PyDict_SetItemString(modules, "yt.utilities.lib.primitives", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "yt/utilities/lib/primitives.pyx":17
 * cimport cython
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * cimport cython.floating
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/primitives.pyx":26
 * 
 * 
 * cdef np.float64_t DETERMINANT_EPS = 1.0e-10             # <<<<<<<<<<<<<<
 * cdef np.float64_t INF = np.inf
 * 
 */
  __pyx_v_2yt_9utilities_3lib_10primitives_DETERMINANT_EPS = 1.0e-10;

  /* "yt/utilities/lib/primitives.pyx":27
 * 
 * cdef np.float64_t DETERMINANT_EPS = 1.0e-10
 * cdef np.float64_t INF = np.inf             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "platform_dep.h" nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_inf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_2yt_9utilities_3lib_10primitives_INF = __pyx_t_3;

  /* "yt/utilities/lib/primitives.pyx":1
 * # distutils: libraries = STD_LIBS             # <<<<<<<<<<<<<<
 * """
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "vec3_ops.pxd":59
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef inline cython.floating L2_norm(const cython.floating[3] a) nogil:             # <<<<<<<<<<<<<<
 *     return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init yt.utilities.lib.primitives", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init yt.utilities.lib.primitives");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int64 neg_one = (npy_int64) -1, const_zero = (npy_int64) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 2 * PyLong_SHIFT) {
                            return (npy_int64) (((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 3 * PyLong_SHIFT) {
                            return (npy_int64) (((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 4 * PyLong_SHIFT) {
                            return (npy_int64) (((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) ((((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int64) -1;
        }
    } else {
        npy_int64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int64) -1;
        val = __Pyx_PyInt_As_npy_int64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int64");
    return (npy_int64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int64");
    return (npy_int64) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int64 neg_one = (npy_int64) -1, const_zero = (npy_int64) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int64),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
